# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow: Visualizations of quantum circuits,
"""

import os
import re
import subprocess
import tempfile
from typing import Any, Dict, List

import sympy
from PIL import Image
from sympy import Symbol

from . import utils, var
from .circuits import Circuit
from .dagcircuit import DAGCircuit
from .gates import P0, P1
from .modules import IdentityGate
from .ops import Gate, Operation
from .qubits import Qubits
from .stdgates import CZ, CSwap, Swap
from .stdops import Reset

__all__ = (
    "LATEX_GATESET",
    "circuit_to_latex",
    "latex_to_image",
    "circuit_to_image",
    "circuit_to_diagram",
)


# TODO: Should be set of types to match GATESET in stdgates?
LATEX_GATESET = frozenset(
    [
        "I",
        "X",
        "Y",
        "Z",
        "H",
        "T",
        "S",
        "T_H",
        "S_H",
        "V",
        "V_H",
        "Rx",
        "Ry",
        "Rz",
        "SqrtY",
        "SqrtY_H",
        "XPow",
        "YPow",
        "ZPow",
        "HPow",
        "CNot",
        "CZ",
        "Swap",
        "ISwap",
        "PSwap",
        "CV",
        "CV_H",
        "CPhase",
        "CH",
        "Can",
        "CCNot",
        "CSwap",
        "CCZ",
        "CCiX",
        "Deutsch",
        "CCXPow",
        "XX",
        "YY",
        "ZZ",
        "Can",
        "P0",
        "P1",
        "Reset",
        "NoWire",
        "Measure",
        "Ph",
        "ECP",
        "SqrtISwap_H",
        "Barenco",
        "CNotPow",
        "CYPow",
        "CZPow",
        "B",
        "CY",
        "ECP",
    ]
)


kwarg_to_symbol = {
    "alpha": Symbol("α"),
    "lam": Symbol("λ"),
    "nx": Symbol("n_x"),
    "ny": Symbol("n_y"),
    "nz": Symbol("n_z"),
    "p": Symbol("p"),
    "phi": Symbol("φ"),
    "t": Symbol("t"),
    "t0": Symbol("t_0"),
    "t1": Symbol("t_1"),
    "t2": Symbol("t_2"),
    "theta": Symbol("θ"),
    "tx": Symbol("t_x"),
    "ty": Symbol("t_y"),
    "tz": Symbol("t_z"),
    "s": Symbol("s"),
    "b": Symbol("b"),
    "c": Symbol("c"),
}
"""Mapping of standard gate arguments to sympy Symbols"""


class NoWire(IdentityGate):
    """Dummy gate used to draw a gap in a circuit"""

    _diagram_labels = ["  "]


def circuit_to_latex(
    circ: Circuit,
    qubits: Qubits = None,
    document: bool = True,
    package: str = "quantikz",
    options: str = None,
    scale: float = 0.75,
    qubit_labels: bool = True,
) -> str:
    """
    Create an image of a quantum circuit in LaTeX.

    Can currently draw X, Y, Z, H, T, S, T_H, S_H, Rx, Ry, Rz, XPow, YPow, ZPow,
    HPow, CNot, CZ, Swap, ISwap, CCNot, CSwap, XX, YY, ZZ, Can, P0 and P1 gates,
    and the Reset operation.

    Args:
        circ:       A quantum Circuit
        qubits:     Optional qubit list to specify qubit order
        document:   If false, just the qcircuit latex is returned. Else the
                    circuit image is wrapped in a standalone LaTeX document
                    ready for typesetting.
        package:    The LaTeX package used for rendering. Either 'qcircuit'
                    or 'quantikz' (default)
    Returns:
        A LaTeX string representation of the circuit.

    Raises:
        NotImplementedError: For unsupported gates.

    Refs:
        LaTeX Qcircuit package
            (https://arxiv.org/pdf/quant-ph/0406003).
        LaTeX quantikz package
            (https://arxiv.org/abs/1809.03842).
    """

    assert package in ["qcircuit", "quantikz"]  # FIXME. Throw Exception

    # TODO: I would be good to move much of the gate dependent details
    # into the gate classes, as has been done for circuit_to_diagram.
    # But there seems to be too many exceptions to be practical. (??)
    from .config import CTRL, TARGET

    latex_labels = {
        "S_H": [r"S^\dagger"],
        "V_H": [r"V^\dagger"],
        "T_H": [r"T^\dagger"],
        "SqrtY": [r"Y^{{\frac{{1}}{{2}}}}"],
        "SqrtY_H": [r"Y^{{-\frac{{1}}{{2}}}}"],
        "Rx": [r"R_x({theta})"],
        "Ry": [r"R_y({theta})"],
        "Rz": [r"R_z({theta})"],
        "XPow": [r"X^{{{t}}}"],
        "YPow": [r"Y^{{{t}}}"],
        "ZPow": [r"Z^{{{t}}}"],
        "HPow": [r"H^{{{t}}}"],
        "XX": [r"X\!X^{{{t}}}"],
        "YY": [r"Y\!Y^{{{t}}}"],
        "ZZ": [r"Z\!Z^{{{t}}}"],
        "ISwap": [r"\text{{iSwap}}"],
        "B": [r"\text{{B}}"],
        "SqrtISwap": [r"\sqrt{{\text{{iSwap}}}}"],
        "SqrtISwap_H": [r"\sqrt{{\text{{iSwap}}}}^\dagger"],
        "CNot": [CTRL, TARGET],
        "CH": [CTRL, "H"],
        "CV": [CTRL, "V"],
        "CV_H": [CTRL, r"V^\dagger"],
        "CNotPow": [CTRL, r"X^{{{t}}}"],
        "CY": [CTRL, r"Y"],
        "CYPow": [CTRL, r"Y^{{{t}}}"],
        "CZPow": [CTRL, r"Z^{{{t}}}"],
        # 3-qubit gates
        "CCNot": [CTRL, CTRL, TARGET],
        "CCXPow": [CTRL, CTRL, r"X^{{{t}}}"],
        "CCZ": [CTRL, CTRL, r"Z"],
        "CCiX": [CTRL, CTRL, r"iX"],
        "Deutsch": [CTRL, CTRL, r"iR^2_x({theta})"],
        "Barenco": [CTRL, r"\text{{Bar}}({phi}, {alpha}, {theta})"],
    }

    if len(circ) == 0:
        # Empty circuit
        circ = Circuit(NoWire([0]))

    if qubits is None:
        qubits = circ.qubits
    N = len(qubits)
    qubit_idx = dict(zip(qubits, range(N)))
    layers = _display_layers(circ, qubits)

    layer_code = []

    if qubit_labels:
        code = [r"\lstick{%s}" % q for q in qubits]
        layer_code.append(code)

    for layer in layers:
        code = [r"\qw"] * N

        for gate in layer:
            elem = gate  # FIXME: elem not gate
            idx = [qubit_idx[q] for q in gate.qubits]

            name = gate.name

            if name not in LATEX_GATESET:
                raise NotImplementedError(str(gate))

            pretty_params: Dict[str, str] = {}
            # FIXME: Do I need gate isinstance (Also below)
            if isinstance(elem, Gate) and elem.params:
                pretty_params = {
                    key: _pretty(value, format="latex")
                    for key, value in zip(gate.cv_args, elem.params)
                }

            # Construct text labels
            name = elem.name

            if name in latex_labels:
                text_labels = latex_labels[name]
                if len(idx) != 1 and len(text_labels) == 1:
                    text_labels = text_labels * len(idx)
                if pretty_params:
                    text_labels = [t.format(**pretty_params) for t in text_labels]
            else:
                if len(name) > 1:
                    name = r"\text{" + name + "}"
                if pretty_params:
                    params_text = ",".join(pretty_params.values())

                    text_labels = [name + "(%s)" % params_text] * len(idx)
                else:
                    text_labels = [name] * len(idx)
                if len(idx) != 1 and not elem.cv_interchangeable:
                    # If not interchangeable, we have to label connections
                    for i in range(elem.qubit_nb):
                        text_labels[i] = text_labels[i] + "_{%s}" % i

            # Special 1-qubit gates
            if isinstance(gate, NoWire):
                if package == "qcircuit":
                    for i in idx:
                        code[i] = r"\push{ }"
                else:  # quantikz
                    for i in idx:
                        code[i] = r""
            elif isinstance(gate, P0):
                code[idx[0]] = r"\push{\ket{0}\!\!\bra{0}} \qw"
            elif isinstance(gate, P1):
                code[idx[0]] = r"\push{\ket{1}\!\!\bra{1}} \qw"
            # elif isinstance(gate, Measure):
            #     code[idx[0]] = r"\meter{}"  # TODO: Add cbit label        # FIXME

            # Special 2-qubit gates
            elif isinstance(gate, CZ):
                code[idx[0]] = r"\ctrl{" + str(idx[1] - idx[0]) + "}"
                code[idx[1]] = r"\ctrl{" + str(idx[0] - idx[1]) + "}"
            elif isinstance(gate, Swap):
                if package == "qcircuit":
                    code[idx[0]] = r"\qswap \qwx[" + str(idx[1] - idx[0]) + "]"
                    code[idx[1]] = r"\qswap"
                else:  # quantikz
                    code[idx[0]] = r"\swap{" + str(idx[1] - idx[0]) + "}"
                    code[idx[1]] = r"\targX{}"

            # interchangeable 2-qubit gate
            elif (
                gate.qubit_nb == 2
                and gate.cv_interchangeable
                and gate.name != "CZPow"
                # and gate.name not in latex_labels
            ):
                top = min(idx)
                bot = max(idx)

                # TODO: either qubits neighbors, in order,
                # or bit symmetric gate
                if package == "qcircuit":
                    if bot - top == 1:
                        code[top] = r"\multigate{1}{%s}" % text_labels[0]
                        code[bot] = r"\ghost{%s}" % text_labels[0]
                    else:
                        code[top] = r"\sgate{%s}{%s}" % (text_labels[0], str(bot - top))
                        code[bot] = r"\gate{%s}" % (text_labels[1])
                else:  # quantikz
                    if bot - top == 1:
                        code[top] = r"\gate[2]{%s}" % text_labels[0]
                        code[bot] = r""
                    else:
                        code[top] = r"\gate{%s}\vqw{%s}" % (
                            text_labels[0],
                            str(bot - top),
                        )
                        code[bot] = r"\gate{%s}" % (text_labels[1])

            # Special three-qubit gates
            elif isinstance(gate, CSwap):
                if package == "qcircuit":
                    code[idx[0]] = r"\ctrl{" + str(idx[1] - idx[0]) + "}"
                    code[idx[1]] = r"\qswap \qwx[" + str(idx[2] - idx[1]) + "]"
                    code[idx[2]] = r"\qswap"
                else:  # quantikz
                    # FIXME: Leaves a weird gap in circuit
                    # code[idx[0]] = r'\ctrl{}\vqw{' + str(idx[1]-idx[0]) + '}'
                    code[idx[0]] = r"\ctrl{" + str(idx[1] - idx[0]) + "}"
                    code[idx[1]] = r"\swap{" + str(idx[2] - idx[1]) + "}"
                    code[idx[2]] = r"\targX{}"

            # Special multi-qubit gates
            elif isinstance(gate, Reset):
                for i in idx:
                    code[i] = r"\push{\rule{0.1em}{0.5em}\, \ket{0}\,} \qw"

            #  Other gates with explicit gate labels
            elif name in latex_labels:
                for i in range(gate.qubit_nb):
                    if text_labels[i] == CTRL:
                        code[idx[i]] = r"\ctrl{" + str(idx[i + 1] - idx[i]) + "}"
                    elif text_labels[i] == TARGET:
                        code[idx[i]] = r"\targ{}"
                    else:
                        code[idx[i]] = r"\gate{" + text_labels[i] + "}"

            # Generic 1-qubit gate
            elif gate.qubit_nb == 1:
                code[idx[0]] = r"\gate{" + text_labels[0] + "}"

            else:
                raise NotImplementedError(str(gate))

        layer_code.append(code)

    code = [r"\qw"] * N
    layer_code.append(code)

    latex_lines = [""] * N

    for line, wire in enumerate(zip(*layer_code)):
        latex = "& " + " & ".join(wire)
        if line < N - 1:  # Not last line
            latex += r" \\"
        latex_lines[line] = latex

    if package == "qcircuit":
        # TODO: qcircuit options
        latex_code = _QCIRCUIT % "\n".join(latex_lines)
        if document:
            latex_code = _QCIRCUIT_HEADER + latex_code + _QCIRCUIT_FOOTER

    else:  # quantikz
        if options is None:
            options = _QUANTIKZ_OPTIONS
        latex_code = _QUANTIKZ % (options, "\n".join(latex_lines))
        latex_code = (r"\adjustbox{scale=%s}{" % scale) + latex_code + "}"

        if document:
            latex_code = _QUANTIKZ_HEADER + latex_code + QUANTIKZ_FOOTER_
    return latex_code


_QCIRCUIT_HEADER = r"""
\documentclass[border={20pt 4pt 20pt 4pt}]{standalone}
\usepackage[braket, qm]{qcircuit}
\usepackage{amsmath}
\usepackage{adjustbox}
\begin{document}
"""

_QCIRCUIT = r"""\Qcircuit @C=1.5em @R=1.5em {
%s
}"""

_QCIRCUIT_FOOTER = r"""
\end{document}
"""

_QUANTIKZ_HEADER = r"""
\documentclass[border={20pt 4pt 20pt 4pt}]{standalone}
\usepackage{amsmath}
\usepackage{tikz}
\usetikzlibrary{quantikz}
\usepackage{adjustbox}
\begin{document}
"""

_QUANTIKZ = r"""\begin{quantikz}[%s]
%s
\end{quantikz}
"""

_QUANTIKZ_OPTIONS = "thin lines, column sep=0.75em," "row sep={2.5em,between origins}"

QUANTIKZ_FOOTER_ = r"""
\end{document}
"""


# TODO: Handle unicode via xelatex?


def latex_to_image(latex: str) -> Image:  # pragma: no cover
    """
    Convert a single page LaTeX document into an image.

    To display the returned image, `img.show()`


    Required external dependencies: `pdflatex` (with `qcircuit` package),
    and `poppler` (for `pdftocairo`).

    Args:
        A LaTeX document as a string.

    Returns:
        A PIL Image

    Raises:
        OSError: If an external dependency is not installed.
    """
    tmpfilename = "circ"
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmppath = os.path.join(tmpdirname, tmpfilename)
        with open(tmppath + ".tex", "w") as latex_file:
            latex_file.write(latex)

        subprocess.run(
            [
                "pdflatex",
                "-halt-on-error",
                f"-output-directory={tmpdirname}",
                f"{tmpfilename}.tex",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            check=True,
        )

        subprocess.run(
            ["pdftocairo", "-singlefile", "-png", "-q", tmppath + ".pdf", tmppath]
        )
        img = Image.open(tmppath + ".png")

    return img


def circuit_to_image(circ: Circuit, qubits: Qubits = None) -> Image:  # pragma: no cover
    """Create an image of a quantum circuit.

    A convenience function that calls circuit_to_latex() and latex_to_image().

    Args:
        circ:       A quantum Circuit
        qubits:     Optional qubit list to specify qubit order

    Returns:
        Returns: A PIL Image (Use img.show() to display)

    Raises:
        NotImplementedError: For unsupported gates.
        OSError: If an external dependency is not installed.
    """
    latex = circuit_to_latex(circ, qubits)
    img = latex_to_image(latex)
    return img


def circuit_to_diagram(
    circ: Circuit,
    qubits: Qubits = None,
    use_unicode: bool = True,
    transpose: bool = False,
    qubit_labels: bool = True,
) -> str:
    """
    Draw a text diagram of a quantum circuit.

    Args:
        circ:           A quantum Circuit
        qubits:         Optional qubit list to specify qubit order
        use_unicode:    If false, return ascii
        qubit_labels:   If false, do not display qubit names

    Returns:
        A string representation of the circuit.
    """
    # Kudos: Inspired by the circuit diagram drawer from cirq
    # https://github.com/quantumlib/Cirq/blob/master/cirq/circuits/circuit.py#L1435

    BOX_CHARS = STD_BOX_CHARS

    def qpad(lines: List[str]) -> List[str]:
        max_length = max(len(k) for k in lines)
        tot_length = max_length + 3

        for k, line in enumerate(lines):
            pad_char = [BOX_CHARS[RIGHT + LEFT], " "][k % 2]
            lines[k] = lines[k].ljust(tot_length, pad_char)
        return lines

    def draw_line(code: List[str], i0: int, i1: int, left_pad: int = 0) -> None:
        i0, i1 = sorted([i0, i1])
        for i in range(i0 + 1, i1, 2):
            code[i] = (" " * left_pad) + BOX_CHARS[BOT + TOP]
        for i in range(i0 + 2, i1, 2):
            code[i] = (BOX_CHARS[LEFT + RIGHT] * left_pad) + BOX_CHARS[CROSS]

    print("Here len", len(circ))
    if len(circ) == 0:
        # Empty circuit
        circ = Circuit(NoWire([0]))

    if qubits is None:
        qubits = circ.qubits
    N = len(qubits)
    # print("Here N", N)

    qubit_idx = dict(zip(qubits, range(0, 2 * N - 1, 2)))
    layers = _display_layers(circ, qubits)
    # print(layers)
    layer_text = []

    qubit_layer = [""] * (2 * N - 1)
    if qubit_labels:
        for n in range(N):
            qubit_layer[n * 2] = str(qubits[n]) + ": "
        max_length = max(len(k) for k in qubit_layer)
        qubit_layer = [line.ljust(max_length) for line in qubit_layer]
    layer_text.append(qpad(qubit_layer))

    for layer in layers:
        code = [""] * (2 * N - 1)

        for elem in layer:
            idx = [qubit_idx[q] for q in elem.qubits]

            # Pretty print parameters
            pretty_params: Dict[str, str] = {}
            params = elem.params
            if params:
                pretty_params = {
                    key: _pretty(value, format="text")
                    for key, value in zip(elem.cv_args, params)
                }

            # Construct text labels
            name = elem.name
            if elem._diagram_labels:
                text_labels = elem._diagram_labels
                if len(idx) != 1 and len(text_labels) == 1:
                    text_labels = list(text_labels) * len(idx)
                text_labels = [t.format(**pretty_params) for t in text_labels]
            else:
                if pretty_params:
                    params_text = ",".join(pretty_params.values())
                    text_labels = [name + "(%s)" % params_text] * len(idx)
                else:
                    text_labels = [name] * len(idx)
                if len(idx) != 1 and not elem.cv_interchangeable:
                    # If not interchangeable, we have to label connections
                    for i in range(elem.qubit_nb):
                        text_labels[i] = text_labels[i] + "_%s" % i

            if not use_unicode:
                text_labels = [_unicode_to_ascii(tl) for tl in text_labels]

            if elem.qubit_nb != 1 and not elem._diagram_noline:
                pad = len(re.split(r"[_^(]+", text_labels[0])[0]) // 2
                draw_line(code, min(idx), max(idx), left_pad=pad)

            for i in range(elem.qubit_nb):
                code[idx[i]] = text_labels[i]
        # end loop over elements

        layer_text.append(qpad(code))
    # end loop over layers

    circ_text = "\n".join("".join(line) for line in zip(*layer_text))

    if transpose:
        boxtrans = dict(zip(BOX_CHARS, _box_char_transpose(BOX_CHARS)))
        circ_text = "".join(boxtrans.get(c, c) for c in circ_text)
        lines = [list(line) for line in circ_text.splitlines()]
        circ_text = "\n".join("".join(c) for c in zip(*lines))

    if not use_unicode:
        circ_text = _unicode_to_ascii(circ_text)

    return circ_text + "\n"


# ==== UTILITIES ====


def _display_layers(circ: Circuit, qubits: Qubits) -> Circuit:
    """Separate a circuit into groups of gates that do not visually overlap"""

    N = len(qubits)
    qubit_idx = dict(zip(qubits, range(N)))

    # Split circuit into Moments, where the elements in each
    # moment operate on non-overlapping qubits
    gate_layers = DAGCircuit(circ).moments()
    # print(len(gate_layers))

    # Now split each moment into visual layers where the
    # control lines do not visually overlap.
    layers = []
    for gl in gate_layers:
        lcirc: List[Operation] = []
        layers.append(lcirc)
        unused = [True] * N
        for gate in gl:
            indices = [qubit_idx[q] for q in gate.qubits]

            if not all(unused[min(indices) : max(indices) + 1]):
                # New layer
                lcirc = []
                layers.append(lcirc)
                unused = [True] * N

            unused[min(indices) : max(indices) + 1] = [False] * (
                max(indices) - min(indices) + 1
            )
            lcirc += gate

    # Sometimes the last layer from one moment can be merged with
    # the first layer of the next moment.
    # FIXME
    # for i in range(len(layers)-1, 1, -1):
    #     qbs0 = set(layers[i-1].qubits)
    #     qbs1 = set(layers[i].qubits)
    #     if qbs0.isdisjoint(qbs1):
    #         print(">>>>>>>>")
    #         layers[i-1].extend(layers[i])
    #         del layers[i]

    return Circuit([Circuit(L) for L in layers])


def _pretty(obj: Any, format: str = "text") -> str:
    """Pretty format an object as a text or latex string."""
    assert format in ["latex", "text"]
    if isinstance(obj, float):
        try:
            if format == "latex":
                return sympy.latex(var.asexpression(obj))
            else:
                return str(var.asexpression(obj)).replace("pi", "π")
        except ValueError:
            return f"{obj:.4g}"

    out = str(obj)
    if format == "text":
        out = out.replace("pi", "π")
    return out


# Unicode and ASCII characters for drawing boxes.
#                  t 0000000011111111
#                  r 0000111100001111
#                  b 0011001100110011
#                  l 0101010101010101
STD_BOX_CHARS = " ╴╷┐╶─┌┬╵┘│┤└┴├┼"  # noqa: E221
BOLD_BOX_CHARS = " ╸╻┓╺━┏┳╹┛┃┫┗┻┣╋"  # noqa: E221
DOUBLE_BOX_CHARS = " ═║╗══╔╦║╝║╣╚╩╠╬"  # noqa: E221  # No half widths
ASCII_BOX_CHARS = r"   \ -/+ /|+\+++"  # noqa: E221

TOP, RIGHT, BOT, LEFT = 8, 4, 2, 1
CROSS = TOP + RIGHT + BOT + LEFT


def _box_char_transpose(chars: str) -> str:
    return "".join(
        chars[utils.bitlist_to_int(list(reversed(utils.int_to_bitlist(n, 4))))]
        for n in range(16)
    )


# FIXME: pi, alpha, ect...
unicode_ascii = {
    "●": "@",
    "○": "O",
    "⁺": "^-1",
    "√": "Sqrt",
    "⟨": "<",  # "Mathematical Left Angle Bracket"
    "⟩": ">",
}
unicode_ascii.update(dict(zip(STD_BOX_CHARS, ASCII_BOX_CHARS)))


def _unicode_to_ascii(text: str) -> str:
    return "".join(unicode_ascii.get(c, c) for c in text)


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


from itertools import chain
from typing import List, Type

import numpy as np
import pytest

import quantumflow as qf
from quantumflow.translate import translation_source_gate
from quantumflow.visualization import kwarg_to_symbol


@pytest.mark.parametrize("trans", qf.TRANSLATORS.values())  # type: ignore
def test_translators(trans: Type[qf.StdGate]) -> None:
    gatet = translation_source_gate(trans)

    args = [np.random.uniform(-4, 4) for _ in gatet.cv_args]
    qbs = range(10, 10 + gatet.cv_qubit_nb)  # Check that qubits are preserved
    gate = gatet(*chain(args, qbs))

    circ1 = qf.Circuit(trans(gate))  # type: ignore
    print(type(circ1[0]))
    circ1[0].tensor
    circ1.asgate()

    circ1.asgate().tensor

    assert qf.gates_close(gate, circ1.asgate())

    # # FIXME. Many translations currently do not respect phase
    # print('Checking gates phase close...')
    # assert qf.gates_phase_close(gate, circ1.asgate())


concrete = {n: np.random.uniform(-4, 4) for n in kwarg_to_symbol.values()}


@pytest.mark.parametrize("trans", qf.TRANSLATORS.values())  # type: ignore
def test_translators_symbolic(trans: Type[qf.StdGate]) -> None:
    """Check that translations can handle symbolic arguments"""
    gatet = translation_source_gate(trans)
    args = [kwarg_to_symbol[a] for a in gatet.cv_args]
    qbs = range(gatet.cv_qubit_nb)
    gate = gatet(*chain(args, qbs))

    qubits = "abcdefg"[0 : gate.qubit_nb]  # Check that qubits are preserved
    gate = gate.on(*qubits)

    circ0 = qf.Circuit([gate])
    circ1 = qf.Circuit(trans(gate))  # type: ignore

    circ0f = circ0.resolve(concrete)
    circ1f = circ1.resolve(concrete)
    assert qf.gates_close(circ0f.asgate(), circ1f.asgate())


def test_translate() -> None:
    circ0 = qf.Circuit([qf.CSwap(0, 1, 2)])

    translators = [
        qf.translate_cswap_to_ccnot,
        qf.translate_ccnot_to_cnot,
        qf.translate_cnot_to_cz,
    ]
    circ1 = qf.circuit_translate(circ0, translators)
    assert circ1.size() == 33

    circ1 = qf.circuit_translate(circ0, translators, recurse=False)

    qf.gates_close(circ0.asgate(), circ1.asgate())

    circ2 = qf.circuit_translate(circ0)
    qf.gates_close(circ0.asgate(), circ2.asgate())


def test_circuit_translate_targets() -> None:
    circ0 = qf.Circuit([qf.CSwap(0, 1, 2)])
    targets = [qf.Can, qf.XPow, qf.ZPow, qf.I]
    circ1 = qf.circuit_translate(circ0, targets=targets)  # type: ignore
    qf.gates_close(circ0.asgate(), circ1.asgate())


def test_can_to_cnot() -> None:
    gate = qf.Can(0.3, 0.23, 0.22, 0, 1)
    circ = qf.Circuit(qf.translate_can_to_cnot(gate))  # type: ignore
    assert qf.gates_close(gate, circ.asgate())

    gate = qf.Can(0.3, 0.23, 0.0, 0, 1)
    circ = qf.Circuit(qf.translate_can_to_cnot(gate))  # type: ignore
    print(qf.canonical_decomposition(circ.asgate()))
    assert qf.gates_close(gate, circ.asgate())


terminal_2q_gate = (
    qf.Can,
    qf.XX,
    qf.YY,
    qf.ZZ,
    qf.CNot,
    qf.CZ,
)


@pytest.mark.parametrize("term_gate", terminal_2q_gate)
def test_decompose_to_terminal_2q_gate(term_gate: Type[qf.StdGate]) -> None:
    # 1 qubit terminal gates
    # We include identity and global phase because translators can't
    # delete gates
    gates = {qf.XPow, qf.ZPow, qf.I, qf.Ph}

    gates.add(term_gate)
    trans = qf.select_translators(gates, qf.TRANSLATORS.values())  # type: ignore
    for t in trans:
        gatet = translation_source_gate(t)
        gates.add(gatet)

    missing = set(qf.StdGate.cv_stdgates.values()) - gates
    if len(missing) != 0:
        print("Missing gates:", missing)
        assert len(missing) == 0


def test_circuit_translate_exception() -> None:
    circ0 = qf.Circuit([qf.CSwap(0, 1, 2)])
    with pytest.raises(ValueError):
        gates: List[Type[qf.StdGate]] = [qf.XPow, qf.ZPow, qf.I, qf.Ph]
        translators = [
            qf.translate_cswap_to_ccnot,
            qf.translate_ccnot_to_cnot,
            qf.translate_cnot_to_cz,
        ]
        qf.circuit_translate(circ0, translators, gates)


# fin



# Copyright 2020-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow: Three qubit gates
"""

import numpy as np

from .. import tensors, utils, var
from ..config import CTRL, SWAP_TARGET, TARGET
from ..ops import StdGate
from ..paulialgebra import Pauli, sX, sZ
from ..qubits import Qubit
from ..states import State
from ..tensors import QubitTensor
from ..var import PI, Variable
from .stdgates_2q import CZ, CNot, Swap

# 3-qubit gates, in alphabetic order

__all__ = ("CCiX", "CCNot", "CCXPow", "CCZ", "CSwap", "Deutsch")


class CCiX(StdGate):
    r"""
    A doubly controlled iX gate.

    Equivalent to ``controlled_gate(cnot())``

    .. math::
        \text{CCiX}() \equiv \begin{pmatrix}
                1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 0 & i \\
                0 & 0 & 0 & 0 & 0 & 0 & i & 0
            \end{pmatrix}

    Refs:
          http://arxiv.org/abs/1210.0974
    """
    # Kudos: Adapted from 1210.0974, via Quipper

    cv_tensor_structure = "monomial"
    _diagram_labels = ("─" + CTRL + "─", "─" + CTRL + "─", "iX─")

    def __init__(self, q0: Qubit = 0, q1: Qubit = 1, q2: Qubit = 2) -> None:
        super().__init__(qubits=[q0, q1, q2])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1, q2 = self.qubits
        return -sX(q2) * (1 - sZ(q1)) * (1 - sZ(q0)) * PI / 8

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = np.asarray(
            [
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0j],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0j, 0.0],
            ]
        )
        return tensors.asqutensor(unitary)


# end class CCiX


class CCNot(StdGate):
    r"""
    A 3-qubit Toffoli gate. A controlled, controlled-not.

    Equivalent to ``controlled_gate(cnot())``

    .. math::
        \text{CCNot}() \equiv \begin{pmatrix}
                1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 \\
                0 & 0 & 0 & 0 & 0 & 0 & 1 & 0
            \end{pmatrix}

    """
    cv_tensor_structure = "permutation"
    _diagram_labels = (CTRL, CTRL, TARGET)

    def __init__(self, q0: Qubit = 0, q1: Qubit = 1, q2: Qubit = 2) -> None:
        super().__init__(qubits=[q0, q1, q2])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1, q2 = self.qubits
        return CNot(q1, q2).hamiltonian * (1 - sZ(q0)) / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 1, 0],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CCNot":
        return self  # Hermitian

    def __pow__(self, t: Variable) -> "CCXPow":
        return CCXPow(t, *self.qubits)

    def run(self, ket: State) -> State:
        axes = ket.qubit_indices(self.qubits)
        s110 = utils.multi_slice(axes, [1, 1, 0])
        s111 = utils.multi_slice(axes, [1, 1, 1])
        tensor = ket.tensor.copy()
        tensor[s110] = ket.tensor[s111]
        tensor[s111] = ket.tensor[s110]
        return State(tensor, ket.qubits, ket.memory)


# end class CCNot


class CCXPow(StdGate):
    r"""
    Powers of the Toffoli gate.

    args:
        t:  turns (powers of the CNOT gate, or controlled-powers of X)
        q0: control qubit
        q1: control qubit
        q2: target qubits
    """
    _diagram_labels = (CTRL, CTRL, "X^{t}")

    def __init__(
        self, t: Variable, q0: Qubit = 0, q1: Qubit = 1, q2: Qubit = 2
    ) -> None:
        super().__init__(params=[t], qubits=[q0, q1, q2])

    @property
    def hamiltonian(self) -> Pauli:
        return CCNot(*self.qubits).hamiltonian * self.param("t")

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = np.pi * self.float_param("t")
        phase = np.exp(0.5j * theta)
        cht = np.cos(theta / 2)
        sht = np.sin(theta / 2)
        unitary = [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, phase * cht, phase * -1.0j * sht],
            [0, 0, 0, 0, 0, 0, phase * -1.0j * sht, phase * cht],
        ]

        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CCXPow":
        return self ** -1

    def __pow__(self, e: Variable) -> "CCXPow":
        (t,) = self.params
        return CCXPow(e * t, *self.qubits)


# end class CCXPow


class CSwap(StdGate):
    r"""
    A 3-qubit Fredkin gate. A controlled swap.

    Equivalent to ``controlled_gate(swap())``

    .. math::
        \text{CSwap}() \equiv \begin{pmatrix}
                1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 0 & 1
            \end{pmatrix}
    """
    cv_hermitian = True
    cv_tensor_structure = "permutation"
    _diagram_labels = (CTRL, SWAP_TARGET, SWAP_TARGET)

    def __init__(self, q0: Qubit = 0, q1: Qubit = 1, q2: Qubit = 2) -> None:
        super().__init__(qubits=[q0, q1, q2])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1, q2 = self.qubits
        return Swap(q1, q2).hamiltonian * (1 - sZ(q0)) / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CSwap":
        return self  # Hermitian

    # TODO: __pow__

    def run(self, ket: State) -> State:
        axes = ket.qubit_indices(self.qubits)
        s101 = utils.multi_slice(axes, [1, 0, 1])
        s110 = utils.multi_slice(axes, [1, 1, 0])
        tensor = ket.tensor.copy()
        tensor[s101] = ket.tensor[s110]
        tensor[s110] = ket.tensor[s101]
        return State(tensor, ket.qubits, ket.memory)


# end class CSWAP


class CCZ(StdGate):
    r"""
    A controlled, controlled-Z.

    Equivalent to ``controlled_gate(CZ())``

    .. math::
        \text{CSWAP}() \equiv \begin{pmatrix}
                1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 0 & -1
            \end{pmatrix}
    """
    cv_interchangeable = True
    cv_tensor_structure = "diagonal"
    _diagram_labels = (CTRL, CTRL, CTRL)

    def __init__(self, q0: Qubit = 0, q1: Qubit = 1, q2: Qubit = 2) -> None:
        super().__init__(qubits=[q0, q1, q2])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1, q2 = self.qubits
        return CZ(q1, q2).hamiltonian * (1 - sZ(q0)) / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, -1],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CCZ":
        return self  # Hermitian

    # TODO: __pow__

    def run(self, ket: State) -> State:
        axes = ket.qubit_indices(self.qubits)
        s11 = utils.multi_slice(axes, [1, 1, 1])
        tensor = ket.tensor.copy()
        tensor[s11] *= -1
        return State(tensor, ket.qubits, ket.memory)


# end class CCZ


# TODO: hamiltonian
class Deutsch(StdGate):
    r"""
    The Deutsch gate, a 3-qubit universal quantum gate.

    A controlled-controlled-i*R_x(2*theta) gate.
    Note that Deutsch(pi/2) is the CCNot gate.

    .. math::
        \text{Deutsch}(\theta) \equiv \begin{pmatrix}
                1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & i \cos(\theta) & \sin(\theta) \\
                0 & 0 & 0 & 0 & 0 & 0 & \sin(\theta)& i \cos(\theta)
            \end{pmatrix}

    Ref:
        D. Deutsch, Quantum Computational Networks,
            Proc. R. Soc. Lond. A 425, 73 (1989).
    """
    _diagram_labels = (CTRL, CTRL, "iRx({theta})^2")

    def __init__(
        self, theta: Variable, q0: Qubit = 0, q1: Qubit = 1, q2: Qubit = 2
    ) -> None:
        super().__init__(params=[theta], qubits=[q0, q1, q2])

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))

        unitary = [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 1.0j * np.cos(theta), np.sin(theta)],
            [0, 0, 0, 0, 0, 0, np.sin(theta), 1.0j * np.cos(theta)],
        ]
        return tensors.asqutensor(unitary)

    # TODO: Specializes to Toffoli


# end class Deutsch



# Copyright 2020-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow: Gates peculiar to Rigetti's Forest
"""

import numpy as np

from .. import tensors, var
from ..config import CTRL, NCTRL
from ..ops import StdGate
from ..paulialgebra import Pauli, sZ
from ..qubits import Qubit
from ..tensors import QubitTensor
from ..utils import cached_property
from ..var import Variable

__all__ = ("CPhase", "CPhase00", "CPhase01", "CPhase10", "PSwap")


class CPhase(StdGate):
    r"""A 2-qubit 11 phase-shift gate

    .. math::
        \text{CPhase}(\theta) \equiv \text{diag}(1, 1, 1, e^{i \theta})
    """
    cv_interchangeable = True
    cv_tensor_structure = "diagonal"
    _diagram_labels = [CTRL + "({theta})", CTRL + "({theta})"]

    def __init__(self, theta: Variable, q0: Qubit = 0, q1: Qubit = 1) -> None:
        super().__init__(params=[theta], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        theta = self.param("theta")
        return -theta * (1 + sZ(q0) * sZ(q1) - sZ(q0) - sZ(q1)) / 4

    @cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        unitary = [
            [1.0, 0, 0, 0],
            [0, 1.0, 0, 0],
            [0, 0, 1.0, 0],
            [0, 0, 0, np.exp(1j * theta)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CPhase":
        return self ** -1

    def __pow__(self, t: Variable) -> "CPhase":
        theta = self.param("theta") * t
        return CPhase(theta, *self.qubits)


# end class CPhase


class CPhase00(StdGate):
    r"""A 2-qubit 00 phase-shift gate

    .. math::
        \text{CPhase00}(\theta) \equiv \text{diag}(e^{i \theta}, 1, 1, 1)
    """
    cv_interchangeable = True
    cv_tensor_structure = "diagonal"
    _diagram_labels = [NCTRL + "({theta})", NCTRL + "({theta})"]

    def __init__(self, theta: Variable, q0: Qubit = 0, q1: Qubit = 1) -> None:
        super().__init__(params=[theta], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        theta = self.param("theta")
        return -theta * (1 + sZ(q0) * sZ(q1) + sZ(q0) + sZ(q1)) / (4)

    @cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        unitary = [
            [np.exp(1j * theta), 0, 0, 0],
            [0, 1.0, 0, 0],
            [0, 0, 1.0, 0],
            [0, 0, 0, 1.0],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CPhase00":
        return self ** -1

    def __pow__(self, t: Variable) -> "CPhase00":
        theta = self.param("theta")
        return CPhase00(theta * t, *self.qubits)


# end class CPhase00


class CPhase01(StdGate):
    r"""A 2-qubit 01 phase-shift gate

    .. math::
        \text{CPhase01}(\theta) \equiv \text{diag}(1, e^{i \theta}, 1, 1)
    """
    cv_tensor_structure = "diagonal"
    _diagram_labels = [NCTRL + "({theta})", CTRL + "({theta})"]

    def __init__(self, theta: Variable, q0: Qubit = 0, q1: Qubit = 1) -> None:
        super().__init__(params=[theta], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        return -self.param("theta") * (1 - sZ(q0) * sZ(q1) + sZ(q0) - sZ(q1)) / (4)

    @cached_property
    def tensor(self) -> QubitTensor:
        unitary = [
            [1.0, 0, 0, 0],
            [0, np.exp(1j * var.asfloat(self.param("theta"))), 0, 0],
            [0, 0, 1.0, 0],
            [0, 0, 0, 1.0],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CPhase01":
        return self ** -1

    def __pow__(self, t: Variable) -> "CPhase01":
        theta = self.param("theta")
        return CPhase01(theta * t, *self.qubits)


# end class CPhase01


class CPhase10(StdGate):
    r"""A 2-qubit 10 phase-shift gate

    .. math::
        \text{CPhase10}(\theta) \equiv \text{diag}(1, 1, e^{i \theta}, 1)
    """
    cv_tensor_structure = "diagonal"
    _diagram_labels = [CTRL + "({theta})", NCTRL + "({theta})"]

    def __init__(self, theta: Variable, q0: Qubit = 0, q1: Qubit = 1) -> None:
        super().__init__(params=[theta], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        theta = self.param("theta")
        return -theta * (1 - sZ(q0) * sZ(q1) - sZ(q0) + sZ(q1)) / (4)

    @cached_property
    def tensor(self) -> QubitTensor:
        unitary = [
            [1.0, 0, 0, 0],
            [0, 1.0, 0, 0],
            [0, 0, np.exp(1j * var.asfloat(self.param("theta"))), 0],
            [0, 0, 0, 1.0],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CPhase10":
        return self ** -1

    def __pow__(self, t: Variable) -> "CPhase10":
        theta = self.param("theta")
        return CPhase10(theta * t, *self.qubits)


# end class CPhase10


class PSwap(StdGate):
    r"""A 2-qubit parametric-swap gate, as defined by Quil.
    Interpolates between SWAP (theta=0) and iSWAP (theta=pi/2).

    Locally equivalent to ``CAN(1/2, 1/2, 1/2 - theta/pi)``

    .. math::
        \text{PSwap}(\theta) \equiv \begin{pmatrix} 1&0&0&0 \\
        0&0&e^{i\theta}&0 \\ 0&e^{i\theta}&0&0 \\ 0&0&0&1 \end{pmatrix}
    """
    cv_interchangeable = True
    cv_tensor_structure = "monomial"

    def __init__(self, theta: Variable, q0: Qubit = 0, q1: Qubit = 1) -> None:
        super().__init__(params=[theta], qubits=[q0, q1])

    @cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        unitary = [
            [[[1, 0], [0, 0]], [[0, 0], [np.exp(theta * 1.0j), 0]]],
            [[[0, np.exp(theta * 1.0j)], [0, 0]], [[0, 0], [0, 1]]],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "PSwap":
        theta = self.param("theta")
        theta = 2.0 * np.pi - theta % (2.0 * np.pi)
        return PSwap(theta, *self.qubits)


# end class PSwap


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
.. contents:: :local:
.. currentmodule:: quantumflow.xquirk

Interface between Quirk and QuantumFlow

https://algassert.com/quirk

.. autofunction:: circuit_to_quirk
.. autofunction:: quirk_url
.. autofunction:: open_quirk_webserver
"""
import json
import urllib
import webbrowser
from typing import Dict, List, cast

from .circuits import Circuit

__all__ = "circuit_to_quirk", "quirk_url", "open_quirk_webserver"


# TODOs
# compile to quirk.
# connectivity: non-control multi-qubit gates have to be contiguous.
#
# CCZPow
# CNotPow
# CCXPow
#
# measurement
# Iden
# Permutations:
# Rotate, Reverse, left rotate, right rotate, Interleave, Deinterleave
#
# Displays
#
# P0, P1
#
# Add QUIRK_OPS
# TODO: Docs

quirk_labels: Dict[str, List] = {
    "I": [1],
    "H": ["H"],
    "X": ["X"],
    "Y": ["Y"],
    "Z": ["Z"],
    "V": ["X^½"],
    "V_H": ["X^-½"],
    "SqrtY": ["Y^½"],
    "SqrtY_H": ["Y^-½"],
    "S": ["Z^½"],
    "S_H": ["Z^-½"],
    "T": ["Z^¼"],
    "T_H": ["Z^-¼"],
    "CNot": ["•", "X"],
    "CY": ["•", "Y"],
    "CZ": ["•", "Z"],
    "Swap": ["Swap", "Swap"],
    "CSwap": ["•", "Swap", "Swap"],
    "CCNot": ["•", "•", "X"],
    "CCZ": ["•", "•", "Z"],
}

quirk_formulaic = {
    "Rx": "Rxft",
    "Ry": "Ryft",
    "Rz": "Rzft",
    "XPow": "X^ft",
    "YPow": "Y^ft",
    "ZPow": "Z^ft",
}


def circuit_to_quirk(circ: Circuit) -> str:
    """Convert a QuantumFlow Circuit to a quirk circuit (represented as a
    JSON formatted string).
    """

    # Relabel qubits to consecutive integers
    N = circ.qubit_nb
    circ = circ.on(*range(0, N))

    columns = []
    col: List = [1] * N
    columns.append(col)

    for op in circ:
        qbs = cast(List[int], op.qubits)

        if op.name in quirk_labels:
            labels = quirk_labels[op.name]
            if "•" in col or not all(col[q] == 1 for q in qbs):
                # New column
                col = [1] * N
                columns.append(col)
            for i, q in enumerate(qbs):
                col[q] = labels[i]
        elif op.name in quirk_formulaic:
            q = qbs[0]
            if "•" in col or col[q] != 1:
                # New column
                col = [1] * N
                columns.append(col)
            (p,) = op.params
            col[q] = {"id": quirk_formulaic[op.name], "arg": str(p)}
        else:
            raise ValueError("Cannot convert Operation to Quirk")

    # Remove excess '1's  at end of column
    for col in columns:
        while len(col) > 1 and col[-1] == 1:
            col.pop()

    quirk = {"cols": columns}
    s = json.dumps(quirk, ensure_ascii=False).replace(" ", "")
    return s


def quirk_url(
    quirk: str,
    base_url: str = "https://algassert.com/quirk#circuit={}",
    escape: bool = False,
) -> str:
    """Given a quirk circuit, returns the corresponding quirk web-server URL"""
    if escape:
        quirk = urllib.parse.quote(quirk)
    return base_url.format(quirk)


def open_quirk_webserver(circ: Circuit) -> None:  # pragma: no cover
    """Translate a Circuit to quirk and open the quirk
    webapp in a browser window"""

    quirk = circuit_to_quirk(circ)
    url = quirk_url(quirk, escape=True)
    webbrowser.open(url)


# fin



# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Unit tests for quantumflow.channels
"""
# Kudos: Tests adapted from density branch of reference-qvm by Nick Rubin

from functools import reduce
from operator import add

import numpy as np
import pytest

import quantumflow as qf


def test_transpose_map() -> None:
    # The transpose map is a superoperator that transposes a 1-qubit
    # density matrix. Not physical.
    # quant-ph/0202124

    q0 = 0
    ops = [
        qf.Unitary(np.asarray([[1, 0], [0, 0]]), [q0]),
        qf.Unitary(np.asarray([[0, 0], [0, 1]]), [q0]),
        qf.Unitary(np.asarray([[0, 1], [1, 0]]) / np.sqrt(2), [q0]),
        qf.Unitary(np.asarray([[0, 1], [-1, 0]]) / np.sqrt(2), [q0]),
    ]

    kraus = qf.Kraus(ops, weights=(1, 1, 1, -1))
    rho0 = qf.random_density(1)
    rho1 = kraus.evolve(rho0)

    op0 = rho0.asoperator()
    op1 = rho1.asoperator()
    assert np.allclose(op0.T, op1)

    # The Choi matrix should be same as Swap operator
    choi = kraus.aschannel().choi()
    choi = choi
    assert np.allclose(choi, qf.Swap(0, 2).asoperator())


def test_random_density() -> None:
    rho = qf.random_density(4)
    assert rho.tensor.shape == (2,) * 8


def test_density() -> None:
    ket = qf.random_state(3)
    matrix = qf.tensors.outer(ket.tensor, np.conj(ket.tensor), rank=1)
    qf.Density(matrix)
    qf.Density(matrix, [0, 1, 2])

    with pytest.raises(ValueError):
        qf.Density(matrix, [0, 1, 2, 3])


def test_state_to_density() -> None:
    density = qf.ghz_state(4).asdensity()
    assert list(density.tensor.shape) == [2] * 8

    prob = density.probabilities()
    assert np.isclose(prob[0, 0, 0, 0] - 0.5, 0.0)
    assert np.isclose(prob[0, 1, 0, 0], 0.0)
    assert np.isclose(prob[1, 1, 1, 1] - 0.5, 0.0)

    ket = qf.random_state(3)
    density = ket.asdensity()
    ket_prob = ket.probabilities()
    density_prob = density.probabilities()

    for index, prob in np.ndenumerate(ket_prob):
        assert np.isclose(prob - density_prob[index], 0.0)


def test_purity() -> None:
    density = qf.ghz_state(4).asdensity()
    assert np.isclose(qf.purity(density), 1.0)

    for _ in range(10):
        rho = qf.random_density(4)
        purity = np.real(qf.purity(rho))
        assert purity < 1.0
        assert purity >= 0.0


def test_stdkraus_creation() -> None:
    qf.Damping(0.1, 0)
    qf.Depolarizing(0.1, 0)
    qf.Dephasing(0.1, 0)


def test_stdchannels_creation() -> None:
    qf.Damping(0.1, 0).aschannel()
    qf.Depolarizing(0.1, 0).aschannel()
    qf.Dephasing(0.1, 0).aschannel()


def test_identity() -> None:
    chan = qf.IdentityGate([1]).aschannel()
    rho = qf.random_density(2)
    after = chan.evolve(rho)
    assert qf.densities_close(rho, after)

    assert chan.name == "Channel"


def test_channel_chi() -> None:
    chan = qf.IdentityGate([0, 1, 2]).aschannel()
    chi = chan.chi()
    assert chi.shape == (64, 64)


def test_channle_choi() -> None:
    chan0 = qf.Damping(0.1, 0).aschannel()
    choi = chan0.choi()
    chan1 = qf.Channel.from_choi(choi, [0])
    assert qf.channels_close(chan0, chan1)


def test_sample_coin() -> None:
    chan = qf.H(0).aschannel()
    rho = qf.zero_state(1).asdensity()
    rho = chan.evolve(rho)
    prob = rho.probabilities()
    assert np.allclose(prob, [[0.5, 0.5]])


def test_sample_bell() -> None:
    rho = qf.zero_state(2).asdensity()
    chan = qf.H(0).aschannel()
    rho = chan.evolve(rho)
    chan = qf.CNot(0, 1).aschannel()
    rho = chan.evolve(rho)
    prob = rho.probabilities()

    assert np.allclose(prob, [[0.5, 0], [0, 0.5]])


def test_biased_coin() -> None:
    # sample from a 75% head and 25% tails coin
    rho = qf.zero_state(1).asdensity()
    chan = qf.Rx(np.pi / 3, 0).aschannel()
    rho = chan.evolve(rho)
    prob = rho.probabilities()
    assert np.allclose(prob, [0.75, 0.25])


def test_measurement() -> None:
    rho = qf.zero_state(2).asdensity()
    chan = qf.H(0).aschannel()
    rho = chan.evolve(rho)
    rho = qf.Kraus([qf.P0(0), qf.P1(0)]).aschannel().evolve(rho)
    K = qf.Kraus([qf.P0(1), qf.P1(1)])
    chan = K.aschannel()

    rho = qf.Kraus([qf.P0(1), qf.P1(1)]).aschannel().evolve(rho)
    prob = rho.probabilities()
    assert np.allclose(prob, [[0.5, 0], [0.5, 0]])
    assert np.isclose(prob[0, 0] * 2, 1.0)
    assert np.isclose(prob[1, 0] * 2, 1.0)


def test_qaoa() -> None:
    ket_true = [
        0.00167784 + 1.00210180e-05 * 1j,
        0.5 - 4.99997185e-01 * 1j,
        0.5 - 4.99997185e-01 * 1j,
        0.00167784 + 1.00210180e-05 * 1j,
    ]
    rho_true = qf.State(ket_true).asdensity()

    rho = qf.zero_state(2).asdensity()
    rho = qf.Ry(np.pi / 2, 0).aschannel().evolve(rho)
    rho = qf.Rx(np.pi, 0).aschannel().evolve(rho)
    rho = qf.Ry(np.pi / 2, 1).aschannel().evolve(rho)
    rho = qf.Rx(np.pi, 1).aschannel().evolve(rho)
    rho = qf.CNot(0, 1).aschannel().evolve(rho)
    rho = qf.Rx(-np.pi / 2, 1).aschannel().evolve(rho)
    rho = qf.Ry(4.71572463191, 1).aschannel().evolve(rho)
    rho = qf.Rx(np.pi / 2, 1).aschannel().evolve(rho)
    rho = qf.CNot(0, 1).aschannel().evolve(rho)
    rho = qf.Rx(-2 * 2.74973750579, 0).aschannel().evolve(rho)
    rho = qf.Rx(-2 * 2.74973750579, 1).aschannel().evolve(rho)
    assert qf.densities_close(rho, rho_true)


def test_amplitude_damping() -> None:
    rho = qf.zero_state(1).asdensity()
    p = 1.0 - np.exp(-(50) / 15000)
    chan = qf.Damping(p, 0).aschannel()
    rho1 = chan.evolve(rho)
    assert qf.densities_close(rho, rho1)

    rho2 = qf.X(0).aschannel().evolve(rho1)
    rho3 = chan.evolve(rho2)

    expected = qf.Density(
        [[0.00332778 + 0.0j, 0.00000000 + 0.0j], [0.00000000 + 0.0j, 0.99667222 + 0.0j]]
    )
    assert qf.densities_close(expected, rho3)


def test_depolarizing() -> None:
    p = 1.0 - np.exp(-1 / 20)

    chan = qf.Depolarizing(p, 0).aschannel()
    rho0 = qf.Density([[0.5, 0], [0, 0.5]])
    assert qf.densities_close(rho0, chan.evolve(rho0))

    rho0 = qf.random_density(1)
    rho1 = chan.evolve(rho0)
    pr0 = np.real(qf.purity(rho0))
    pr1 = np.real(qf.purity(rho1))
    assert pr0 > pr1

    # Test data extracted from refereneqvm
    rho2 = qf.Density([[0.43328691, 0.48979689], [0.48979689, 0.56671309]])
    rho_test = qf.Density(
        [[0.43762509 + 0.0j, 0.45794666 + 0.0j], [0.45794666 + 0.0j, 0.56237491 + 0.0j]]
    )
    assert qf.densities_close(chan.evolve(rho2), rho_test)

    ket0 = qf.random_state(1)
    qf.Depolarizing(p, 0).run(ket0)

    rho1b = qf.Depolarizing(p, 0).evolve(rho0)
    assert qf.densities_close(rho1, rho1b)


def test_kruas_qubits() -> None:
    rho = qf.Kraus([qf.P0(0), qf.P1(1)])
    assert rho.qubits == (0, 1)
    assert rho.qubit_nb == 2


def test_kraus_evolve() -> None:
    rho = qf.zero_state(1).asdensity()
    p = 1 - np.exp(-50 / 15000)
    kraus = qf.Damping(p, 0)
    rho1 = kraus.evolve(rho)
    assert qf.densities_close(rho, rho1)

    rho2 = qf.X(0).aschannel().evolve(rho1)
    rho3 = kraus.evolve(rho2)

    expected = qf.Density(
        [[0.00332778 + 0.0j, 0.00000000 + 0.0j], [0.00000000 + 0.0j, 0.99667222 + 0.0j]]
    )

    assert qf.densities_close(expected, rho3)


def test_kraus_run() -> None:
    ket0 = qf.zero_state(["a"])
    ket0 = qf.X("a").run(ket0)
    p = 1.0 - np.exp(-2000 / 15000)

    kraus = qf.Damping(p, "a")

    reps = 1000
    results = [kraus.run(ket0).asdensity().asoperator() for _ in range(reps)]
    matrix = reduce(add, results) / reps
    rho_kraus = qf.Density(matrix, ["a"])

    rho0 = ket0.asdensity()
    chan = kraus.aschannel()
    rho_chan = chan.evolve(rho0)

    # If this fails occasionally consider increasing tolerance
    # Can't be very tolerant due to stochastic dynamics

    assert qf.densities_close(rho_chan, rho_kraus, atol=0.05)


def test_channel_adjoint() -> None:
    kraus0 = qf.Damping(0.1, 0)
    chan0 = kraus0.aschannel()
    chan1 = chan0.H.H
    assert qf.channels_close(chan0, chan1)

    chan2 = kraus0.H.aschannel()
    assert qf.channels_close(chan2, chan0.H)

    # 2 qubit hermitian channel
    chan3 = qf.CZ(0, 1).aschannel()
    chan4 = chan3.H
    assert qf.channels_close(chan3, chan4)


def test_kraus_qubits() -> None:
    kraus = qf.Kraus([qf.X(1), qf.Y(0)])
    assert kraus.qubits == (0, 1)

    kraus = qf.Kraus([qf.X("a"), qf.Y("b")])
    assert len(kraus.qubits) == 2
    assert kraus.qubit_nb == 2


def test_chan_qubits() -> None:
    chan = qf.Kraus([qf.X(1), qf.Y(0)]).aschannel()
    assert chan.qubits == (0, 1)
    assert chan.qubit_nb == 2


def test_chan_permute() -> None:
    chan0 = qf.CNot(0, 1).aschannel()
    chan1 = qf.CNot(1, 0).aschannel()

    assert not qf.channels_close(chan0, chan1)

    chan2 = chan1.permute([0, 1])
    assert chan2.qubits == (0, 1)
    assert qf.channels_close(chan1, chan2)

    chan3 = chan1.on(0, 1)
    print(chan0.qubits, chan3.qubits)
    assert qf.channels_close(chan0, chan3)


def test_channel_errors() -> None:
    chan = qf.CNot(0, 1).aschannel()
    with pytest.raises(TypeError):
        chan.run(qf.zero_state(2))

    with pytest.raises(TypeError):
        chan.asgate()

    assert chan.aschannel() is chan

    with pytest.raises(NotImplementedError):
        chan @ 123  # type: ignore

    with pytest.raises(ValueError):
        qf.Channel(chan.tensor, qubits=[0, 1, 2])


def test_kraus_errors() -> None:
    kraus = qf.Kraus([qf.X(1), qf.Y(0)])

    with pytest.raises(TypeError):
        kraus.asgate()


def test_kraus_complete() -> None:
    kraus = qf.Kraus([qf.X(1)])
    assert qf.kraus_iscomplete(kraus)

    kraus = qf.Damping(0.1, 0)
    assert qf.kraus_iscomplete(kraus)

    assert qf.kraus_iscomplete(qf.Damping(0.1, 0))
    assert qf.kraus_iscomplete(qf.Dephasing(0.1, 0))
    assert qf.kraus_iscomplete(qf.Depolarizing(0.1, 0))


def test_askraus() -> None:
    def _roundtrip(kraus: qf.Kraus) -> None:
        assert qf.kraus_iscomplete(kraus)

        chan0 = kraus.aschannel()
        kraus1 = qf.channel_to_kraus(chan0)
        assert qf.kraus_iscomplete(kraus1)

        chan1 = kraus1.aschannel()
        assert qf.channels_close(chan0, chan1)

    p = 1 - np.exp(-50 / 15000)
    _roundtrip(qf.Kraus([qf.X(1)]))
    _roundtrip(qf.Damping(p, 0))
    _roundtrip(qf.Depolarizing(0.9, 1))


def test_channel_trace() -> None:
    chan = qf.I(0).aschannel()
    assert np.isclose(chan.trace(), 4)


def test_channel_join() -> None:
    chan0 = qf.H(0).aschannel()
    chan1 = qf.X(1).aschannel()
    chan01 = qf.join_channels(chan0, chan1)
    assert chan01.qubits == (0, 1)


def test_create_channel() -> None:
    gate = qf.Rx(0.2, 0)
    _ = qf.Rx(0.2, 0).aschannel()

    params = tuple(gate.params)
    qubits = gate.qubits
    tensor = gate.aschannel().tensor
    name = "ChanRx"

    _ = qf.Channel(tensor, qubits, params, name)


def test_random_channel() -> None:
    chan = qf.random_channel([0, 1, 2], rank=4)
    rho0 = qf.random_density([0, 1, 2], rank=2)
    purity0 = qf.purity(rho0)
    rho1 = chan.evolve(rho0)
    purity1 = qf.purity(rho1)

    assert purity1 < purity0

    chan = qf.random_channel([0, 1, 2], rank=4, unital=True)
    assert qf.almost_unital(chan)


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Unit tests for quantumflow examples.
"""

import subprocess

import numpy as np

import quantumflow as qf


def test_prepare_w4():
    import examples.state_prep_w4 as ex

    ket = ex.prepare_w4()
    assert qf.states_close(ket, qf.w_state(4))


def test_prepare_w4_main():
    rval = subprocess.call(["examples/state_prep_w4.py"])
    assert rval == 0


def test_prepare_w16():
    import examples.state_prep_w16 as ex

    ket = ex.prepare_w16()
    assert qf.states_close(ket, qf.w_state(16))


def test_prepare_w16_main():
    rval = subprocess.call(["examples/state_prep_w16.py"])
    assert rval == 0


def test_prepare_cswap_decomposition():
    rval = subprocess.call(["examples/cswap_decomposition.py"])
    assert rval == 0


def test_swap_test():
    import examples.swaptest as ex

    ket0 = qf.zero_state([0])
    ket1 = qf.random_state([1])
    ket2 = qf.random_state([2])

    ket = qf.join_states(ket0, ket1)
    ket = qf.join_states(ket, ket2)

    fid = qf.state_fidelity(ket1, ket2.on(1))
    st_fid = ex.swap_test(ket, 0, 1, 2)

    assert np.isclose(fid, st_fid)


def test_swap_test_main():
    rval = subprocess.call(["examples/swaptest.py"])
    assert rval == 0


def test_circuit_identities_main():
    rval = subprocess.call(["examples/circuit_identities.py"])
    assert rval == 0


def test_gate_translate_identities_main():
    rval = subprocess.call(["examples/gate_translations.py"])
    assert rval == 0


# def test_fit_state():
#     from examples import fit_state
#     fit_state.example_fit_state()

# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Useful routines not necessarily intended to be part of the public API.
"""

import functools
import warnings
from fractions import Fraction
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import networkx as nx
import numpy as np

__all__ = (
    # Future
    "importlib_metadata",
    "cached_property",
    "deprecated",
    # Collections
    "multi_slice",
    "invert_map",
    "FrozenDict",
    # Math
    "almost_integer",
    "bitlist_to_int",
    "int_to_bitlist",
    "rationalize",
    "complex_ginibre_ensemble",
    "unitary_ensemble",
    # Graphs
    "from_graph6",
    "to_graph6",
    "spanning_tree_count",
    "octagonal_tiling_graph",
    "truncated_grid_2d_graph",
)


# -- Future present --

try:
    # python >= 3.8
    from importlib import metadata as importlib_metadata  # type: ignore
except ImportError:  # pragma: no cover
    # python == 3.7
    import importlib_metadata  # type: ignore  # noqa: F401


try:  # pragma: no cover
    # Python >= 3.8
    from functools import cached_property  # type: ignore
except ImportError:  # pragma: no cover
    # python == 3.7
    def cached_property(func):  # type: ignore
        """
        Method decorator for immutable properties.
        The result is cached on the first call.
        """

        def wrapper(instance):  # type: ignore
            attr = "_cached_property_" + func.__name__
            if hasattr(instance, attr):
                return getattr(instance, attr)
            result = func(instance)
            setattr(instance, attr, result)
            return result

        return property(wrapper)


def deprecated(func: Callable) -> Callable:
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def _new_func(*args: Any, **kwargs: Any) -> Any:
        warnings.simplefilter("always", DeprecationWarning)  # turn off filter
        warnings.warn(
            f"Call to deprecated function {func.__name__}.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return _new_func


# -- Collections --


def multi_slice(axes: Sequence, items: Sequence, axes_nb: int = None) -> Tuple:
    """Construct a multidimensional slice to access array data.
    e.g. mydata[multi_slice([3,5], [1,2]))] is equivalent to
    mydata[:,:,:,1,:,2,...]
    """

    N = max(axes) + 1 if axes_nb is None else axes_nb
    slices: List[Any] = [slice(None)] * N
    for axis, item in zip(axes, items):
        slices[axis] = item
    if axes_nb is None:
        slices.append(Ellipsis)
    return tuple(slices)


def invert_map(mapping: dict, one_to_one: bool = True) -> dict:
    """Invert a dictionary. If not one_to_one then the inverted
    map will contain lists of former keys as values.
    """
    if one_to_one:
        inv_map = {value: key for key, value in mapping.items()}
    else:
        inv_map = {}
        for key, value in mapping.items():
            inv_map.setdefault(value, set()).add(key)

    return inv_map


KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


class FrozenDict(Mapping[KeyType, ValueType]):
    """
    An immutable frozen dictionary.

    The FrozenDict is hashable if all the keys and values are hashable.
    """

    # Note: For no apparently good reason, the python standard library contains
    # immutable lists and sets (tuple and frozenset), but does not have an
    # immutable dictionary.

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._dict: Dict[KeyType, ValueType] = dict(*args, **kwargs)
        self._hash: Optional[int] = None

    def __getitem__(self, key: KeyType) -> ValueType:
        return self._dict[key]

    def __contains__(self, key: object) -> bool:
        return key in self._dict

    def copy(self, *args: Any, **kwargs: Any) -> "FrozenDict":
        return self.update()

    def update(self, *args: Any, **kwargs: Any) -> "FrozenDict":
        """Update mappings, and return a new FrizenDict"""
        d = self._dict.copy()
        d.update(*args, **kwargs)
        return type(self)(d)

    def __iter__(self) -> Iterator[KeyType]:
        yield from self._dict

    def __len__(self) -> int:
        return len(self._dict)

    def __repr__(self) -> str:
        return "%s(%r)" % (self.__class__.__name__, self._dict)

    def __hash__(self) -> int:
        if not self._hash:
            self._hash = hash(frozenset(self._dict.items()))
        return self._hash


# -- Math --


def almost_integer(number: Union[float, complex], atol: float = 1e-08) -> bool:
    if isinstance(number, complex):
        if not np.isclose(number.imag, 0, atol=atol):
            return False
        number = number.real
    x = np.isclose(round(number) - number, 0, atol=atol)
    return x


def bitlist_to_int(bitlist: Sequence[int]) -> int:
    """Converts a sequence of bits to an integer.

    >>> from quantumflow.utils import bitlist_to_int
    >>> bitlist_to_int([1, 0, 0])
    4
    """
    return int("".join([str(d) for d in bitlist]), 2)


def int_to_bitlist(x: int, pad: int = None) -> List[int]:
    """Converts an integer to a binary sequence of bits.

    Pad prepends with sufficient zeros to ensures that the returned list
    contains at least this number of bits.

    >>> from quantumflow.utils import int_to_bitlist
    >>> int_to_bitlist(4, 4))
    [0, 1, 0, 0]
    """
    if pad is None:
        form = "{:0b}"
    else:
        form = "{:0" + str(pad) + "b}"

    return list(int(b) for b in form.format(x))


_DENOMINATORS = set(
    [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        12,
        16,
        32,
        64,
        100,
        128,
        256,
        512,
        1000,
        1024,
        2048,
        4096,
        8192,
        10000,
    ]
)


def rationalize(flt: float, denominators: Set[int] = None) -> Fraction:
    """Convert a floating point number to a Fraction with a small
    denominator.

    Args:
        flt:            A floating point number
        denominators:   Collection of standard denominators. Default is
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 16, 32, 64, 100, 128, 256, 512,
            1000, 1024, 2048, 4096, 8192, 10000

    Raises:
        ValueError:     If cannot rationalize float
    """
    if denominators is None:
        denominators = _DENOMINATORS
    frac = Fraction.from_float(flt).limit_denominator()
    if frac.denominator not in denominators:
        raise ValueError("Cannot rationalize")
    return frac


def complex_ginibre_ensemble(size: Tuple[int, ...]) -> np.ndarray:
    """Returns a random complex matrix with values drawn from a standard normal
    distribution.

    Ref:
        Ginibre, Jean (1965). "Statistical ensembles of complex, quaternion,
        and real matrices". J. Math. Phys. 6: 440–449.
        doi:10.1063/1.1704292.
    """
    return np.random.normal(size=size) + 1j * np.random.normal(size=size)


def unitary_ensemble(dim: int) -> np.ndarray:
    """Return a random unitary of size (dim, dim) drawn from Harr measure

    Ref:
       "How to generate random matrices from the classical compact groups",
        Francesco Mezzadri, Notices Am. Math. Soc. 54, 592 (2007).
        arXiv:math-ph/0609050
    """
    import scipy

    return scipy.stats.unitary_group.rvs(dim)


# -- Graphs --


def from_graph6(graph6: str) -> nx.Graph:
    """Convert a string encoded in graph6 format to a networkx graph"""
    return nx.from_graph6_bytes(bytes(graph6, "utf-8"))


def to_graph6(graph: nx.Graph) -> str:
    """Convert a networkx graph to a string in graph6 format"""
    # Networkx makes this surprisingly tricky.
    return nx.to_graph6_bytes(graph)[10:-1].decode("utf-8")


def spanning_tree_count(graph: nx.Graph) -> int:
    """Return the number of unique spanning trees of a graph, using
    Kirchhoff's matrix tree theorem.
    """
    laplacian = nx.laplacian_matrix(graph).toarray()
    comatrix = laplacian[:-1, :-1]
    det = np.linalg.det(comatrix)
    count = int(round(det))
    return count


def octagonal_tiling_graph(M: int, N: int) -> nx.Graph:
    """Return the octagonal tiling graph (4.8.8, truncated square tiling,
    truncated quadrille) of MxNx8 nodes

    To visualize with networkx and matplot lib:
    > import matplotlib.pyplot as plt
    > nx.draw(G, pos={node: node for node in G.nodes})
    >plt.show()
    """

    grp = nx.Graph()
    octogon = [
        [(1, 0), (0, 1)],
        [(0, 1), (0, 2)],
        [(0, 2), (1, 3)],
        [(1, 3), (2, 3)],
        [(2, 3), (3, 2)],
        [(3, 2), (3, 1)],
        [(3, 1), (2, 0)],
        [(2, 0), (1, 0)],
    ]
    left = [[(1, 0), (1, -1)], [(2, 0), (2, -1)]]
    up = [[(0, 1), (-1, 1)], [(0, 2), (-1, 2)]]

    for m in range(M):
        for n in range(N):
            edges = octogon
            if n != 0:
                edges = edges + left
            if m != 0:
                edges = edges + up

            for (x0, y0), (x1, y1) in edges:
                grp.add_edge((m * 4 + x0, n * 4 + y0), (m * 4 + x1, n * 4 + y1))
    return grp


def truncated_grid_2d_graph(m: int, n: int, t: int = None) -> nx.Graph:
    """Generate a rectangular grid graph (of width `m` and height `n`),
    with corners removed. It the truncation `t` is not given, then it
    is set to half the shortest side (rounded down)

    truncated_grid_graph(12, 11) returns the topology of Google's
    bristlecone chip.
    """
    G = nx.grid_2d_graph(m, n)
    if t is None:
        t = min(m, n) // 2

    for mm in range(t):
        for nn in range(t - mm):
            G.remove_node((mm, nn))
            G.remove_node((m - mm - 1, nn))
            G.remove_node((mm, n - nn - 1))
            G.remove_node((m - mm - 1, n - nn - 1))
    return G


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import random

import numpy as np

import quantumflow as qf

from ..config_test import REPS


def test_XPow() -> None:
    assert qf.gates_close(qf.V(0), qf.XPow(0.5, 0))
    assert qf.gates_close(qf.V_H(0), qf.XPow(-0.5, 0))
    assert qf.gates_close(qf.V(1).H, qf.V_H(1))
    assert qf.gates_close(qf.V(1).H.H, qf.V(1))
    assert qf.gates_close(qf.V(0) ** 0.5, qf.XPow(0.25, 0))
    assert qf.gates_close(qf.V_H(0) ** 0.5, qf.XPow(-0.25, 0))
    assert qf.gates_close(qf.V(0) @ qf.V(0), qf.X(0))


def test_rn() -> None:
    theta = 1.23

    gate = qf.Rn(theta, 1, 0, 0, "q0")
    assert qf.gates_close(gate, qf.Rx(theta, "q0"))

    gate = qf.Rn(theta, 0, 1, 0, "q0")
    assert qf.gates_close(gate, qf.Ry(theta, "q0"))

    gate = qf.Rn(theta, 0, 0, 1, "q0")
    assert qf.gates_close(gate, qf.Rz(theta, "q0"))

    gate = qf.Rn(np.pi, 1 / np.sqrt(2), 0, 1 / np.sqrt(2), "q0")
    assert qf.gates_close(gate, qf.H("q0"))


def test_parametric_Z() -> None:
    assert qf.gates_close(qf.ZPow(0.25, 1), qf.T(1))
    assert qf.gates_close(qf.ZPow(0.5, 2), qf.S(2))
    assert qf.gates_close(qf.ZPow(1.0, 3), qf.Z(3))


def test_hadamard() -> None:
    gate = qf.Rz(np.pi / 2, 0) @ qf.I(0)
    gate = qf.Rx(np.pi / 2, 0) @ gate
    gate = qf.Rz(np.pi / 2, 0) @ gate

    assert qf.gates_close(gate, qf.H(0))


def test_rotation_gates() -> None:
    assert qf.gates_close(qf.I(0), qf.I(0))
    assert qf.gates_close(qf.Rx(np.pi, 0), qf.X(0))
    assert qf.gates_close(qf.Ry(np.pi, 0), qf.Y(0))
    assert qf.gates_close(qf.Rz(np.pi, 0), qf.Z(0))


def test_phaseshift() -> None:
    gate = qf.T(0) @ qf.T(0)

    assert qf.gates_close(gate, gate)

    assert qf.gates_close(gate, qf.S(0))
    assert qf.gates_close(qf.S(0), qf.PhaseShift(np.pi / 2, 0))
    assert qf.gates_close(qf.T(0), qf.PhaseShift(np.pi / 4, 0))

    # PhaseShift and RZ are the same up to a global phase.
    for _ in range(REPS):
        theta = random.uniform(-4 * np.pi, +4 * np.pi)
        assert qf.gates_close(qf.Rz(theta, 0), qf.PhaseShift(theta, 0))

    # Causes a rounding error that can result in NANs if not corrected for.
    theta = -2.5700302313621375
    assert qf.gates_close(qf.Rz(theta, 0), qf.PhaseShift(theta, 0))


def test_tz_specialize() -> None:
    for t in [-0.25, 0, 0.25, 0.5, 1.0, 1.5, 1.75, 2.0]:
        gate0 = qf.ZPow(t, 0)
        gate1 = gate0.specialize()
        assert qf.gates_close(gate0, gate1)
        assert not isinstance(gate1, qf.ZPow)


def test_specialize_1q() -> None:
    q0 = 10
    assert isinstance(qf.ZPow(0.0, q0).specialize(), qf.I)
    assert isinstance(qf.ZPow(0.25, q0).specialize(), qf.T)
    assert isinstance(qf.ZPow(0.5, q0).specialize(), qf.S)
    assert isinstance(qf.ZPow(1.0, q0).specialize(), qf.Z)
    assert isinstance(qf.ZPow(1.5, q0).specialize(), qf.S_H)
    assert isinstance(qf.ZPow(1.75, q0).specialize(), qf.T_H)
    assert isinstance(qf.ZPow(1.99999999999, q0).specialize(), qf.I)

    special_values = (
        0.0,
        0.25,
        0.5,
        1.0,
        1.5,
        1.75,
        2.0,
        0.17365,
        np.pi,
        np.pi / 2,
        np.pi / 4,
    )
    special1p1q = (
        qf.Rx,
        qf.Ry,
        qf.Rz,
        qf.XPow,
        qf.YPow,
        qf.ZPow,
        qf.PhaseShift,
        qf.HPow,
        qf.W,
    )

    for gatetype in special1p1q:
        for value in special_values:
            gate0 = gatetype(value, "q0")
            gate1 = gate0.specialize()
            assert gate0.qubits == gate1.qubits
            assert qf.gates_close(gate0, gate1)

    assert isinstance(qf.PhasedXPow(0.1234, 0.0, 1).specialize(), qf.I)
    assert isinstance(qf.PhasedXPow(0.0, 1.0, 1).specialize(), qf.X)
    assert isinstance(qf.PhasedXPow(0.0213, 1.0, 1).specialize(), qf.PhasedXPow)


# fin



# Copyright 2019-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Unit tests for quantumflow.xqiskit
"""

import pytest

pytest.importorskip("qiskit")

from qiskit import (  # noqa: E402
    ClassicalRegister,
    QuantumCircuit,
    QuantumRegister,
)

import quantumflow as qf  # noqa: E402
from quantumflow.xqiskit import QiskitSimulator  # noqa: E402
from quantumflow.xqiskit import circuit_to_qiskit  # noqa: E402
from quantumflow.xqiskit import qiskit_to_circuit  # noqa: E402
from quantumflow.xqiskit import translate_gates_to_qiskit  # noqa: E402


def test_qiskit_to_circuit() -> None:

    q = QuantumRegister(5)
    c = ClassicalRegister(5)
    qc = QuantumCircuit(q, c)

    qc.ccx(q[0], q[1], q[2])
    qc.ch(q[0], q[1])
    qc.crz(0.1, q[0], q[1])
    qc.cswap(q[0], q[1], q[2])

    # The QuantumCircuit.cu1 method is deprecated as of 0.16.0.
    # You should use the QuantumCircuit.cp method instead, which acts identically.

    # The QuantumCircuit.cu3 method is deprecated as of 0.16.0.
    # You should use the QuantumCircuit.cu method instead,
    # where cu3(ϴ,φ,λ) = cu(ϴ,φ,λ,0).

    # The QuantumCircuit.u1 method is deprecated as of 0.16.0.
    # You should use the QuantumCircuit.p method instead, which acts identically.

    # The QuantumCircuit.u2 method is deprecated as of 0.16.0.
    # You can use the general 1-qubit gate QuantumCircuit.u instead:
    # u2(φ,λ) = u(π/2, φ, λ).
    # Alternatively, you can decompose it in terms of QuantumCircuit.p
    # and QuantumCircuit.sx
    # u2(φ,λ) = p(π/2+φ) sx p(π/2+λ) (1 pulse on hardware).

    # The QuantumCircuit.u3 method is deprecated as of 0.16.0.
    # You should use QuantumCircuit.u instead, which acts identically.
    # Alternatively, you can decompose u3 in terms of QuantumCircuit.p
    # and QuantumCircuit.sx
    # u3(ϴ,φ,λ) = p(φ+π) sx p(ϴ+π) sx p(λ) (2 pulses on hardware).

    qc.cu1(0.1, q[0], q[1])
    qc.cu3(0.1, 0.2, 0.3, q[0], q[1])
    qc.cx(q[0], q[1])
    qc.cy(q[0], q[1])
    qc.cz(q[0], q[1])
    qc.h(q[0])
    qc.i(q[1])
    qc.i(q[2])
    qc.rx(0.0, q[0])
    qc.ry(0.1, q[1])
    qc.rz(0.2, q[2])
    qc.rzz(0.1, q[0], q[1])
    qc.s(q[2])
    qc.sdg(q[2])
    qc.swap(q[0], q[1])
    qc.t(q[1])
    qc.tdg(q[1])
    qc.u1(0.2, q[2])
    qc.u2(0.1, 0.2, q[2])
    qc.u3(0.1, 0.2, 0.3, q[2])
    qc.x(q[0])
    qc.y(q[0])
    qc.z(q[0])

    circ = qiskit_to_circuit(qc)
    # print(circ)

    assert (
        str(circ)
        == """Circuit
    CCNot 0 1 2
    CH 0 1
    CRZ(1/10) 0 1
    CSwap 0 1 2
    CPhase(1/10) 0 1
    CU3(1/10, 1/5, 3/10) 0 1
    CNot 0 1
    CY 0 1
    CZ 0 1
    H 0
    I 1
    I 2
    Rx(0) 0
    Ry(1/10) 1
    Rz(1/5) 2
    RZZ(1/10) 0 1
    S 2
    S_H 2
    Swap 0 1
    T 1
    T_H 1
    PhaseShift(1/5) 2
    U2(1/10, 1/5) 2
    U3(1/10, 1/5, 3/10) 2
    X 0
    Y 0
    Z 0"""
    )

    circuit_to_qiskit(circ)


def test_qiskit_if() -> None:
    q = QuantumRegister(5)
    c = ClassicalRegister(5)
    qc = QuantumCircuit(q, c)

    qc.x(q[2]).c_if(c, 1)

    circ = qiskit_to_circuit(qc)
    op = circ[0]
    assert isinstance(op, qf.If)
    assert op.element.name == "X"
    assert op.key == c
    assert op.value == 1


def test_circuit_to_qiskit() -> None:
    circ = qf.Circuit()
    circ += qf.X(0)
    circ += qf.Y(1)
    circ += qf.Z(2)
    circ += qf.Can(0.1, 0.2, 0.2, 0, 1)

    circ1 = translate_gates_to_qiskit(circ)
    print()
    print(qf.circuit_to_diagram(circ1))

    qc = circuit_to_qiskit(circ, translate=True)
    print(qc)

    assert len(circ1) == len(qc)


def test_qiskitsimulator() -> None:
    circ = qf.Circuit()
    circ += qf.H(1)
    circ += qf.X(0)
    circ += qf.H(2)
    circ += qf.Y(3)
    circ += qf.Z(2)
    circ += qf.Can(0.1, 0.2, 0.2, 0, 1)

    sim = QiskitSimulator(circ)
    assert qf.states_close(circ.run(), sim.run())

    ket0 = qf.random_state([0, 1, 2, 3])
    assert qf.states_close(circ.run(ket0), sim.run(ket0))


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


"""
.. contents:: :local:
.. currentmodule:: quantumflow


QuantumFlow supports several different quantum operations that act upon either
pure or mixed states (or both). The four main types are Gate, which represents
the action of an operator (typically unitary) on a state; Channel, which
represents the action of a superoperator on a mixed state (used for mixed
quantum-classical dynamics); Kraus, which represents a Channel as a collection
of operators; and Circuit, which is a list of other operations that act in
sequence. Circuits can contain any instance of the abstract quantum operation
superclass, Operation, including other circuits.

Quantum operations are immutable, and transformations of these operations return
new copies. (An exception is the composite operation DAGCircuit.)

The main types of Operation's are Gate, Unitary, StdGate, Channel, Circuit, DAGCircuit,
and Pauli.

.. autoclass:: Operation
    :members:

"""

from abc import ABC  # Abstract Base Class
from copy import copy
from typing import (
    Any,
    ClassVar,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

import numpy as np
from scipy.linalg import fractional_matrix_power as matpow
from scipy.linalg import logm

from . import tensors, utils, var
from .qubits import Qubit, Qubits
from .states import Density, State
from .tensors import QubitTensor, TensorLike
from .var import Variable

__all__ = ["Operation", "Gate", "StdGate", "Unitary", "Channel"]


OperationType = TypeVar("OperationType", bound="Operation")
"""Generic type annotations for subtypes of Operation"""

GateType = TypeVar("GateType", bound="Gate")
"""Generic type annotations for subtypes of Gate"""


class Operation(ABC):
    """An operation on a quantum state. An element of a quantum circuit.

    Abstract Base Class for Gate, Circuit, and other quantum operations.

    Attributes:
        qubits: The qubits that this Operation acts upon.
        params: Optional keyword parameters used to create this gate
    """

    # Note: We prefix static class variables with "cv_" to avoid confusion
    # with instance variables

    __slots__ = ["_tensor", "_qubits", "_params"]

    cv_interchangeable: ClassVar[bool] = False
    """Is this a multi-qubit operation that is known to be invariant under
    permutations of qubits?"""

    cv_qubit_nb: ClassVar[int] = None
    """The number of qubits, for operations with a fixed number of qubits"""

    cv_args: Tuple[str, ...] = ()
    """The names of the parameters for this operation"""

    _diagram_labels: ClassVar[Optional[Sequence[str]]] = None
    """Override default labels for drawing text circuit diagrams.
    See visualizations.circuit_to_diagram()"""

    _diagram_noline: ClassVar[bool] = False
    """Override default to not draw a line between qubit wires for multi-qubit
    operations. See visualizations.circuit_to_diagram()"""

    def __init__(
        self,
        qubits: Qubits,
        params: Sequence[Variable] = None,
    ) -> None:
        self._qubits: Qubits = tuple(qubits)
        self._params: Tuple[Variable, ...] = ()
        if params is not None:
            self._params = tuple(params)
        self._tensor: QubitTensor = None

        if self.cv_qubit_nb is not None:
            if self.cv_qubit_nb != len(self._qubits):
                raise ValueError(
                    "Wrong number of qubits for Operation"
                )  # pragma: no cover

    def __iter__(self) -> Iterator["Operation"]:
        yield self

    @property
    def name(self) -> str:
        """Return the name of this Operation"""
        return type(self).__name__

    @property
    def qubits(self) -> Qubits:
        """Return the total number of qubits"""
        return self._qubits

    @property
    def qubit_nb(self) -> int:
        """Return the total number of qubits"""
        return len(self.qubits)

    def on(self: OperationType, *qubits: Qubit) -> OperationType:
        """Return a copy of this Operation with new qubits"""
        if len(qubits) != self.qubit_nb:
            raise ValueError("Wrong number of qubits")
        op = copy(self)
        op._qubits = qubits
        return op

    def rewire(self: OperationType, labels: Dict[Qubit, Qubit]) -> OperationType:
        """Relabel qubits and return copy of this Operation"""
        qubits = tuple(labels[q] for q in self.qubits)
        return self.on(*qubits)

    def qubit_indices(self, qubits: Qubits) -> List[int]:
        """Convert qubits to index positions.

        Raises:
            ValueError: If argument qubits are not found in operation's qubits
        """
        return [self.qubits.index(q) for q in qubits]

    @property
    def params(self) -> Tuple[Variable, ...]:
        """Return all of the parameters of this Operation"""
        return self._params

    def param(self, name: str) -> Variable:
        """Return a a named parameters of this Operation.

        Raise:
            KeyError: If unrecognized parameter name
        """
        try:
            idx = self.cv_args.index(name)
        except ValueError:
            raise KeyError("Unknown parameter name", name)

        return self._params[idx]

    def float_param(self, name: str, subs: Mapping[str, float] = None) -> float:
        """Return a a named parameters of this Operation as a float.

        Args:
            name: The name of the parameter (should be in cls.cv_args)
            subs: Symbolic substitutions to resolve symbolic Variables
        Raise:
            KeyError:  If unrecognized parameter name
            ValueError: If Variable cannot be converted to float
        """
        return var.asfloat(self.param(name), subs)

    def resolve(self, subs: Mapping[str, float]) -> "Operation":
        """Resolve symbolic parameters"""
        # params = {k: var.asfloat(v, subs) for k, v in self.params.items()}
        op = copy(self)
        _params = [var.asfloat(v, subs) for v in self.params]
        op._params = tuple(_params)
        op._tensor = None
        return op

    def asgate(self) -> "Gate":
        """
        Convert this quantum operation to a gate (if possible).

        Raises:
            ValueError: If this operation cannot be converted to a Gate
        """
        raise ValueError()  # pragma: no cover

    def aschannel(self) -> "Channel":
        """Convert this quantum operation to a channel (if possible).

        Raises:
            ValueError: If this operation cannot be converted to a Channel
        """
        raise ValueError()  # pragma: no cover

    @property
    def H(self) -> "Operation":
        """Return the Hermitian conjugate of this quantum operation.

        For unitary Gates (and Circuits composed of the same) the
        Hermitian conjugate returns the inverse Gate (or Circuit)

        Raises:
            ValueError: If this operation does not support Hermitian conjugate
        """
        raise ValueError(
            "This operation does not support Hermitian conjugate"
        )  # pragma: no cover

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        """
        Returns the tensor representation of this operation (if possible)
        """
        raise NotImplementedError()

    def run(self, ket: State) -> State:
        """Apply the action of this operation upon a pure state"""
        raise NotImplementedError()

    def evolve(self, rho: Density) -> Density:
        """Apply the action of this operation upon a mixed state"""
        raise NotImplementedError()

    # Make Operations sortable. (So we can use Operations in opt_einsum
    # axis labels.)
    def __lt__(self, other: Any) -> bool:
        return id(self) < id(other)

    def specialize(self) -> "Operation":
        """For parameterized operations, return appropriate special cases
        for particular parameters. Else return the original Operation.

              e.g. Rx(0.0, 0).specialize() -> I(0)
        """
        return self  # pragma: no cover

    def decompose(self) -> Iterator["Operation"]:
        """Decompose this operation into smaller or more standard subunits.
        If cannot be decomposed, returns self.

        Returns: An iteration of operations.
        """
        yield self  # pragma: no cover

    def _repr_png_(self) -> Optional[bytes]:
        """Jupyter/IPython rich display"""
        from .circuits import Circuit

        return Circuit(self)._repr_png_()

    def _repr_html_(self) -> Optional[str]:
        """Jupyter/IPython rich display"""
        from .circuits import Circuit

        return Circuit(self)._repr_html_()


# End class Operation


class Gate(Operation):
    """
    A quantum logic gate. A unitary operator that acts upon a collection
    of qubits.
    """

    cv_hermitian: ClassVar[bool] = False
    """Is this Gate know to always be hermitian?"""

    cv_tensor_structure: ClassVar[Optional[str]] = None
    """
    Is the tensor representation of this Operation known to have a particular
    structure in the computational basis?

    Options:
        identity
        diagonal
        permutation
        monomial

    A permutation matrix permutes states. It has a single '1' in each row and column.
    All other entries are zero.

    A monomial matrix is a product of a diagonal and a permutation matrix.
    Only 1 entry in each row and column is non-zero.

    """
    # TODO: Add "swap" t roster of tensor structures

    # Note: Circular import hell
    from .paulialgebra import Pauli

    @property
    def hamiltonian(self) -> "Pauli":
        """
        Returns the Hermitian Hamiltonian of corresponding to this
        unitary operation.

        .. math::
            U = e^{-i H)

        Returns:
            A Hermitian operator represented as an element of the Pauli algebra.
        """
        # See test_gate_hamiltonians()
        from .paulialgebra import pauli_decompose_hermitian

        H = -logm(self.asoperator()) / 1.0j
        pauli = pauli_decompose_hermitian(H, self.qubits)
        return pauli

    def asgate(self: GateType) -> GateType:
        return self

    def aschannel(self) -> "Channel":
        """Convert a Gate into a Channel"""
        N = self.qubit_nb
        R = 4

        # TODO: As Kraus?
        tensor = np.outer(self.tensor, self.H.tensor)
        tensor = np.reshape(tensor, [2 ** N] * R)
        tensor = np.transpose(tensor, [0, 3, 1, 2])

        return Channel(tensor, self.qubits)

    def __pow__(self, t: float) -> "Gate":
        """Return this gate raised to the given power."""
        matrix = matpow(self.asoperator(), t)
        return Unitary(matrix, self.qubits)

    def permute(self, qubits: Qubits) -> "Gate":
        """Permute the order of the qubits"""
        if self.qubits == qubits:
            return self
        if self.cv_interchangeable:
            return self.on(*qubits)
        tensor = tensors.permute(self.tensor, self.qubit_indices(qubits))
        return Unitary(tensor, qubits)

    def asoperator(self) -> QubitTensor:
        """Return tensor with with qubit indices flattened"""
        return tensors.flatten(self.tensor, rank=2)
        # N = self.qubit_nb
        # return np.reshape(self.tensor, [2**N] * 2)

    def su(self) -> "Unitary":
        """Convert gate tensor to the special unitary group."""
        rank = 2 ** self.qubit_nb
        U = self.asoperator()
        U /= np.linalg.det(U) ** (1 / rank)
        return Unitary(U, self.qubits)

    @property
    def H(self) -> "Gate":
        return Unitary(self.asoperator().conj().T, self.qubits)

    def __matmul__(self, other: "Gate") -> "Gate":
        """Apply the action of this gate upon another gate,
        self_gate @ other_gate (Note time runs right to left with
        matrix notation)

        Note that the gates must act on the same qubits.
        When gates don't act on the same qubits, use
        Circuit(self_gate, other_gate).asgate() instead.
        """
        if not isinstance(other, Gate):
            raise NotImplementedError()
        gate0 = self
        gate1 = other
        indices = gate1.qubit_indices(gate0.qubits)
        tensor = tensors.tensormul(gate0.tensor, gate1.tensor, tuple(indices))
        return Unitary(tensor, gate1.qubits)

    def run(self, ket: State) -> State:
        """Apply the action of this gate upon a state"""
        if self.cv_tensor_structure == "identity":
            return ket

        qubits = self.qubits
        indices = ket.qubit_indices(qubits)
        tensor = tensors.tensormul(
            self.tensor,
            ket.tensor,
            tuple(indices),
            self.cv_tensor_structure == "diagonal",
        )
        return State(tensor, ket.qubits, ket.memory)

    def evolve(self, rho: Density) -> Density:
        """Apply the action of this gate upon a density"""
        # TODO: implement without explicit channel creation? With Kraus?
        chan = self.aschannel()
        return chan.evolve(rho)

    def specialize(self) -> "Gate":
        return self


# End class Gate


class Unitary(Gate):
    """
    A quantum logic gate specified by an explicit unitary operator.
    """

    def __init__(self, tensor: TensorLike, qubits: Qubits) -> None:

        tensor = tensors.asqutensor(tensor)

        N = np.ndim(tensor) // 2

        if len(tuple(qubits)) != N:
            raise ValueError("Wrong number of qubits for tensor")

        super().__init__(qubits=qubits)
        self._tensor = tensor

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        """Returns the tensor representation of gate operator"""
        return self._tensor


# End class Unitary


# TODO: Move?
class StdGate(Gate):
    """
    A standard gate. Standard gates have a name, a fixed number of real
    parameters, and act upon a fixed number of qubits.

    e.g. Rx(theta, q0), CNot(q0, q1), Can(tx, ty, tz, q0, q1, q2)

    In the argument list, parameters are first, then qubits. Parameters
    have type Variable (either a concrete floating point number, or a symbolic
    expression), and qubits have type Qubit (Any hashable python type).
    """

    cv_stdgates: Dict[str, Type["StdGate"]] = {}
    """List of all StdGate subclasses"""

    def __init_subclass__(cls) -> None:
        # Note: The __init_subclass__ initializes all subclasses of a given class.
        # see https://www.python.org/dev/peps/pep-0487/

        # Parse the Gate arguments and number of qubits from the arguments to __init__
        # Convention is that qubit names start with "q", but arguments do not.
        names = getattr(cls, "__init__").__annotations__.keys()
        args = tuple(s for s in names if s[0] != "q" and s != "return")
        qubit_nb = len(names) - len(args)
        if "return" in names:
            # For unknown reasons, "return" is often, but not always in names.
            qubit_nb -= 1

        cls.cv_args = args
        cls.cv_qubit_nb = qubit_nb

        cls.cv_stdgates[cls.__name__] = cls  # Subclass registration

    def __repr__(self) -> str:
        args: List[str] = []
        args.extend(str(p) for p in self.params)
        args.extend(str(qubit) for qubit in self.qubits)
        fargs = ", ".join(args)

        return f"{self.name}({fargs})"

    def __str__(self) -> str:
        def _param_format(obj: Any) -> str:
            if isinstance(obj, float):
                try:
                    return str(var.asexpression(obj))
                except ValueError:
                    return f"{obj}"
            return str(obj)

        fqubits = " " + " ".join([str(qubit) for qubit in self.qubits])

        if self.params:
            fparams = "(" + ", ".join(_param_format(p) for p in self.params) + ")"
        else:
            fparams = ""

        return f"{self.name}{fparams}{fqubits}"

    def decompose(self) -> Iterator["StdGate"]:
        from .translate import TRANSLATORS, translation_source_gate

        # Terminal gates
        if self.name in ("I", "Ph", "X", "Y", "Z", "XPow", "YPow", "ZPow", "CNot"):
            yield self
            return

        for trans in TRANSLATORS.values():
            from_gate = translation_source_gate(trans)
            if isinstance(self, from_gate):
                yield from trans(self)
                return

        yield self  # fall back  # pragma: no cover


# End class StdGate


class Channel(Operation):
    """A quantum channel"""

    def __init__(
        self,
        tensor: TensorLike,
        qubits: Qubits,
        params: Sequence[var.Variable] = None,
        name: str = None,  # FIXME
    ) -> None:

        tensor = tensors.asqutensor(tensor)

        N = np.ndim(tensor) // 4
        if len(qubits) != N:
            raise ValueError("Wrong number of qubits for tensor")

        super().__init__(qubits=qubits, params=params)
        self._tensor = tensor
        self._name = type(self).__name__ if name is None else name

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        """Return the tensor representation of the channel's superoperator"""
        return self._tensor

    @property
    def name(self) -> str:
        return self._name

    def permute(self, qubits: Qubits) -> "Channel":
        """Return a copy of this channel with qubits in new order"""
        if self.qubits == qubits:
            return self
        # if self.cv_interchangeable:   # TODO
        #     return self.on(*qubits)
        tensor = tensors.permute(self.tensor, self.qubit_indices(qubits))
        return Channel(tensor, qubits=qubits)

    @property
    def H(self) -> "Channel":
        return Channel(tensor=tensors.conj_transpose(self.tensor), qubits=self.qubits)

    # TESTME
    @property
    def sharp(self) -> "Channel":
        r"""Return the 'sharp' transpose of the superoperator.

        The transpose :math:`S^\#` switches the two covariant (bra)
        indices of the superoperator. (Which in our representation
        are the 2nd and 3rd super-indices)

        If :math:`S^\#` is Hermitian, then :math:`S` is a Hermitian-map
        (i.e. transforms Hermitian operators to Hermitian operators)

        Flattening the :math:`S^\#` superoperator to a matrix gives
        the Choi matrix representation. (See channel.choi())
        """

        N = self.qubit_nb

        # TODO: Use tensor_transpose, or remove tensor_transpose
        tensor = self.tensor
        tensor = np.reshape(tensor, [2 ** N] * 4)
        tensor = np.transpose(tensor, (0, 2, 1, 3))
        tensor = np.reshape(tensor, [2] * 4 * N)
        return Channel(tensor, self.qubits)

    def choi(self) -> QubitTensor:
        """Return the Choi matrix representation of this super
        operator"""
        # Put superop axes in the order [out_ket, in_bra, out_bra, in_ket]
        # and reshape to matrix
        N = self.qubit_nb
        return np.reshape(self.sharp.tensor, [2 ** (N * 2)] * 2)

    @classmethod
    def from_choi(cls, tensor: TensorLike, qubits: Qubits) -> "Channel":
        """Return a Channel from a Choi matrix"""
        return cls(tensor, qubits).sharp

    # TESTME
    # FIXME: Can't be right, same as choi?
    def chi(self) -> QubitTensor:
        """Return the chi (or process) matrix representation of this
        superoperator"""
        N = self.qubit_nb
        return np.reshape(self.sharp.tensor, [2 ** (N * 2)] * 2)

    def run(self, ket: State) -> "State":
        raise TypeError()  # Not possible in general

    def evolve(self, rho: Density) -> Density:
        """Apply the action of this channel upon a density"""
        N = rho.qubit_nb
        qubits = rho.qubits

        indices = list([qubits.index(q) for q in self.qubits]) + list(
            [qubits.index(q) + N for q in self.qubits]
        )

        tensor = tensors.tensormul(self.tensor, rho.tensor, tuple(indices))
        return Density(tensor, qubits, rho.memory)

    def asgate(self) -> "Gate":
        raise TypeError()  # Not possible in general

    def aschannel(self) -> "Channel":
        return self

    def __matmul__(self, other: "Channel") -> "Channel":
        """Apply the action of this channel upon another channel,
        self_chan @ other_chan (Note time runs right to left with
        matrix notation)

        Note that the channels must act on the same qubits.
        When gates don't act on the same qubits, use
        Circuit(other_chan, self_chan).aschannel() instead.
        """
        if not isinstance(other, Channel):
            raise NotImplementedError()
        chan0 = self
        chan1 = other
        N = chan1.qubit_nb
        qubits = chan1.qubits
        indices = list([chan1.qubits.index(q) for q in chan0.qubits]) + list(
            [chan1.qubits.index(q) + N for q in chan0.qubits]
        )

        tensor = tensors.tensormul(chan0.tensor, chan1.tensor, tuple(indices))

        return Channel(tensor, qubits)

    # TESTME
    def trace(self) -> QubitTensor:
        """Return the trace of this super operator"""
        return tensors.trace(self.tensor, rank=4)

    # TESTME # TODO
    # def partial_trace(self, qubits: Qubits) -> "Channel":
    #     """Return the partial trace over the specified qubits"""
    #     vec = tensors.partial_trace(self.tensor, qubits, rank=4)
    #     return Channel(vec.tensor, vec.qubits)


# end class Channel

# fin



# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
.. contents:: :local:
.. currentmodule:: quantumflow


Mixed States and Quantum Channels
#################################
.. autoclass:: Density
    :members:

.. autoclass:: Channel
    :members:

.. autoclass:: Kraus
    :members:

.. autoclass:: UnitaryMixture
    :members:

Actions on Densities
####################
.. autofunction:: mixed_density
.. autofunction:: random_density
.. autofunction:: join_densities


Actions on Channels
###################
.. autofunction:: join_channels
.. autofunction:: channel_to_kraus
.. autofunction:: kraus_iscomplete


Standard channels
#################
.. autoclass:: Dephasing
    :members:

.. autoclass:: Damping
    :members:

.. autoclass:: Depolarizing
    :members:

.. autofunction:: random_channel
"""

# Kudos: Kraus maps originally adapted from Nick Rubin's reference-qvm

from functools import reduce
from operator import add
from typing import Sequence

import numpy as np
from scipy import linalg

from . import tensors, utils
from .ops import Channel, Gate, Operation, Unitary
from .qubits import Qubit, Qubits
from .states import Density, State
from .stdgates import I, X, Y, Z

__all__ = [
    "Kraus",
    "UnitaryMixture",
    "Depolarizing",
    "Damping",
    "Dephasing",
    "join_channels",
    "channel_to_kraus",
    "kraus_iscomplete",
    "random_channel",
]


class Kraus(Operation):
    """A Kraus representation of a quantum channel"""

    # DOCME: operator-value-sum representation
    # FIXME: Shouldn't take Gate, since may not be Unitary operators

    def __init__(
        self, operators: Sequence[Gate], weights: Sequence[float] = None
    ) -> None:
        self.operators = operators

        if weights is None:
            weights = [1.0] * len(operators)

        self.weights = tuple(weights)

    def asgate(self) -> Gate:
        """Not possible in general. (But see UnitaryMixture)

        Raises: TypeError
        """
        raise TypeError("Not possible in general")

    def aschannel(self) -> Channel:
        """Returns: Action of Kraus operators as a superoperator Channel"""
        qubits = self.qubits
        N = len(qubits)
        ident = Unitary(np.eye(2 ** N), qubits).aschannel()

        tensors = [(op.aschannel() @ ident).tensor for op in self.operators]
        if self.weights is not None:
            tensors = [t * w for t, w in zip(tensors, self.weights)]
        chan_tensor = reduce(add, tensors)

        return Channel(chan_tensor, self.qubits)

    def run(self, ket: State) -> State:
        """Apply the action of this Kraus quantum operation upon a state"""
        res = [op.run(ket) for op in self.operators]
        probs = [ket.norm() * w for ket, w in zip(res, self.weights)]
        probs = np.asarray(np.abs(probs))
        probs /= np.sum(probs)
        n = np.random.choice(len(res), p=probs)
        newket = res[n]
        return newket.normalize()

    def evolve(self, rho: Density) -> Density:
        """Apply the action of this Kraus quantum operation upon a density"""
        qubits = rho.qubits
        results = [op.evolve(rho) for op in self.operators]
        tensors = [rho.tensor * w for rho, w in zip(results, self.weights)]
        tensor = reduce(add, tensors)
        return Density(tensor, qubits)

    @property
    def qubits(self) -> Qubits:
        """Returns: List of qubits acted upon by this Kraus operation

        The list of qubits is ordered if the qubits labels can be sorted,
        else the the order is indeterminate.

        Raises:
            TypeError: If qubits cannot be sorted into unique order.
        """
        qbs = [q for elem in self.operators for q in elem.qubits]  # gather
        qbs = list(set(qbs))  # unique
        qbs = sorted(qbs)  # sort
        return tuple(qbs)

    @property
    def H(self) -> "Kraus":
        """Return the complex conjugate of this Kraus operation"""
        operators = [op.H for op in self.operators]
        return Kraus(operators, self.weights)


# End class Kraus


class UnitaryMixture(Kraus):
    """A Kraus channel which is a convex mixture of unitary dynamics.

    This Channel is unital, but not all unital channels are unitary
    mixtures.
    """

    def __init__(
        self, operators: Sequence[Gate], weights: Sequence[float] = None
    ) -> None:

        from .info import almost_unitary

        for op in operators:
            if not almost_unitary(op):
                raise ValueError("Operators not all unitary")  # pragma: no cover

        if weights is not None and not np.isclose(np.sum(weights), 1.0):
            raise ValueError("Weights must sum to unity")  # pragma: no cover  # TESTME

        super().__init__(operators, weights)
        # TODO: Sanity check. operators unitary, weights unit

    def asgate(self) -> Gate:
        """Return one of the composite Kraus operators at random with
        the appropriate weights"""
        return np.random.choice(self.operators, p=self.weights)

    def run(self, ket: State) -> State:
        return self.asgate().run(ket)


class Depolarizing(UnitaryMixture):
    """A Kraus representation of a depolarizing channel on 1-qubit.

    Args:
        prob:   The one-step depolarizing probability.
        q0:     The qubit on which to act.
    """

    def __init__(self, prob: float, q0: Qubit) -> None:
        operators = [I(q0), X(q0), Y(q0), Z(q0)]
        weights = [1 - prob, prob / 3.0, prob / 3.0, prob / 3.0]
        super().__init__(operators, weights)


class Damping(Kraus):
    """A Kraus representation of an amplitude-damping (spontaneous emission)
    channel on one qubit

    Args:
        prob:   The one-step damping probability.
        q0:     The qubit on which to act.
    """

    def __init__(self, prob: float, q0: Qubit) -> None:
        kraus0 = Unitary([[1.0, 0.0], [0.0, np.sqrt(1 - prob)]], [q0])
        kraus1 = Unitary([[0.0, np.sqrt(prob)], [0.0, 0.0]], [q0])
        super().__init__([kraus0, kraus1])


class Dephasing(UnitaryMixture):
    """A Kraus representation of a phase-damping quantum channel

    Args:
        prob:   The one-step damping probability.
        q0:     The qubit on which to act.
    """

    def __init__(self, prob: float, q0: Qubit) -> None:
        operators = [I(q0), Z(q0)]
        weights = [1 - prob / 2, prob / 2]
        super().__init__(operators, weights)


def join_channels(chan0: Channel, chan1: Channel) -> Channel:
    """Join two channels acting on different qubits into a single channel
    acting on all qubits"""
    tensor = tensors.outer(chan0.tensor, chan1.tensor, rank=4)
    return Channel(tensor, tuple(chan0.qubits) + tuple(chan1.qubits))


# TESTME
def channel_to_kraus(chan: Channel) -> "Kraus":
    """Convert a channel superoperator into a Kraus operator representation
    of the same channel."""
    qubits = chan.qubits
    N = chan.qubit_nb

    choi = chan.choi()
    evals, evecs = np.linalg.eig(choi)
    evecs = np.transpose(evecs)

    assert np.allclose(evals.imag, 0.0)  # FIXME exception
    assert np.all(evals.real >= 0.0)  # FIXME exception

    values = np.sqrt(evals.real)

    ops = []
    for i in range(2 ** (2 * N)):
        if not np.isclose(values[i], 0.0):
            mat = np.reshape(evecs[i], (2 ** N, 2 ** N)) * values[i]
            g = Unitary(mat, qubits)
            ops.append(g)

    return Kraus(ops)


def kraus_iscomplete(kraus: Kraus) -> bool:
    """Returns True if the collection of (weighted) Kraus operators are
    complete. (Which is necessary for a CPTP map to preserve trace)
    """
    qubits = kraus.qubits
    N = kraus.qubit_nb

    ident = Unitary(np.eye(2 ** N), qubits)

    tensors = [(op.H @ op @ ident).asoperator() for op in kraus.operators]
    tensors = [t * w for t, w in zip(tensors, kraus.weights)]

    tensor = reduce(np.add, tensors)
    res = Unitary(tensor, qubits)

    N = res.qubit_nb
    return np.allclose(res.asoperator(), np.eye(2 ** N))


# TODO: as class RandomChannel?
# Author: GEC (2019)
def random_channel(qubits: Qubits, rank: int = None, unital: bool = False) -> Channel:
    """
    Returns: A randomly sampled Channel drawn from the BCSZ ensemble with
    the specified Kraus rank.

    Args:
        qubits: A list, or number, of qubits.
        rank: Kraus rank of channel. (Defaults to full rank)

    Ref:
        "Random quantum operations", Bruzda, Cappellini, Sommers, and
        Zyczkowski, Physics Letters A 373, 320 (2009). arXiv:0804.2361
    """
    qubits = tuple(qubits)
    N = len(qubits)
    dim = 2 ** N  # Hilbert space dimension
    size = (dim ** 2, dim ** 2) if rank is None else (dim ** 2, rank)

    # arXiv:0804.2361 page 4, steps 1 to 4
    # arXiv:0709.0824 page 6
    X = utils.complex_ginibre_ensemble(size)
    XX = X @ X.conj().T

    if unital:
        Y = np.einsum("ikjk -> ij", XX.reshape([dim, dim, dim, dim]))
        Q = np.kron(linalg.sqrtm(linalg.inv(Y)), np.eye(dim))
    else:
        Y = np.einsum("kikj -> ij", XX.reshape([dim, dim, dim, dim]))
        Q = np.kron(np.eye(dim), linalg.sqrtm(linalg.inv(Y)))

    choi = Q @ XX @ Q

    return Channel.from_choi(choi, qubits)


# Fin



# Copyright 2019-, Gavin E. Crooks and contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import io
from typing import Type

import numpy as np
import pytest

import quantumflow as qf

from .config_test import REPS
from .stdgates.stdgates_test import _randomize_gate


def test_bits() -> None:

    ket = qf.X(0).run(qf.zero_state(4))
    assert ket.tensor[1, 0, 0, 0] == 1
    ket = qf.X(1).run(qf.zero_state(4))
    assert ket.tensor[0, 1, 0, 0] == 1
    ket = qf.X(2).run(qf.zero_state(4))
    assert ket.tensor[0, 0, 1, 0] == 1
    ket = qf.X(3).run(qf.zero_state(4))
    assert ket.tensor[0, 0, 0, 1] == 1

    ket = qf.zero_state(8)
    ket = qf.X(2).run(ket)
    ket = qf.X(4).run(ket)
    ket = qf.X(6).run(ket)

    res = ket.tensor
    assert res[0, 0, 1, 0, 1, 0, 1, 0]


def test_CZ() -> None:
    ket = qf.zero_state(2)
    ket = qf.CZ(0, 1).run(ket)
    assert ket.tensor[0, 0] == 1.0

    ket = qf.X(0).run(ket)
    ket = qf.X(1).run(ket)
    ket = qf.CZ(0, 1).run(ket)
    assert -ket.tensor[1, 1] == 1.0


def test_qaoa_circuit() -> None:
    # Kudos: Adapted from reference QVM
    wf_true = np.array(
        [
            0.00167784 + 1.00210180e-05 * 1j,
            0.50000000 - 4.99997185e-01 * 1j,
            0.50000000 - 4.99997185e-01 * 1j,
            0.00167784 + 1.00210180e-05 * 1j,
        ]
    )
    ket_true = qf.State(wf_true.reshape((2, 2)))

    ket = qf.zero_state(2)
    ket = qf.Ry(np.pi / 2, 0).run(ket)
    ket = qf.Rx(np.pi, 0).run(ket)
    ket = qf.Ry(np.pi / 2, 1).run(ket)
    ket = qf.Rx(np.pi, 1).run(ket)
    ket = qf.CNot(0, 1).run(ket)
    ket = qf.Rx(-np.pi / 2, 1).run(ket)
    ket = qf.Ry(4.71572463191, 1).run(ket)
    ket = qf.Rx(np.pi / 2, 1).run(ket)
    ket = qf.CNot(0, 1).run(ket)
    ket = qf.Rx(-2 * 2.74973750579, 0).run(ket)
    ket = qf.Rx(-2 * 2.74973750579, 1).run(ket)

    assert qf.states_close(ket, ket_true)


def test_qubit_qaoa_circuit() -> None:
    # Adapted from reference QVM
    wf_true = np.array(
        [
            0.00167784 + 1.00210180e-05 * 1j,
            0.50000000 - 4.99997185e-01 * 1j,
            0.50000000 - 4.99997185e-01 * 1j,
            0.00167784 + 1.00210180e-05 * 1j,
        ]
    )
    ket_true = qf.State(wf_true.reshape((2, 2)))

    ket = qf.zero_state(2)
    ket = qf.Ry(np.pi / 2, 0).run(ket)
    ket = qf.Rx(np.pi, 0).run(ket)
    ket = qf.Ry(np.pi / 2, 1).run(ket)
    ket = qf.Rx(np.pi, 1).run(ket)
    ket = qf.CNot(0, 1).run(ket)
    ket = qf.Rx(-np.pi / 2, 1).run(ket)
    ket = qf.Ry(4.71572463191, 1).run(ket)
    ket = qf.Rx(np.pi / 2, 1).run(ket)
    ket = qf.CNot(0, 1).run(ket)
    ket = qf.Rx(-2 * 2.74973750579, 0).run(ket)
    ket = qf.Rx(-2 * 2.74973750579, 1).run(ket)

    assert qf.states_close(ket, ket_true)


@pytest.mark.parametrize("gatet", qf.StdGate.cv_stdgates.values())
def test_gate_run(gatet: Type[qf.StdGate]) -> None:
    gate0 = _randomize_gate(gatet)

    gate1 = qf.Unitary(gate0.tensor, gate0.qubits)
    ket = qf.random_state(gate0.qubits)

    ket0 = gate0.run(ket)
    ket1 = gate1.run(ket)
    assert qf.states_close(ket0, ket1)


@pytest.mark.parametrize("gatet", qf.StdGate.cv_stdgates.values())
def test_gate_evolve(gatet: Type[qf.StdGate]) -> None:
    gate0 = _randomize_gate(gatet)

    gate1 = qf.Unitary(gate0.tensor, gate0.qubits)
    rho = qf.random_density(gate0.qubits)

    rho0 = gate0.evolve(rho)
    rho1 = gate1.evolve(rho)
    assert qf.densities_close(rho0, rho1)


# Test PROJECTORS....
def test_projectors() -> None:
    ket = qf.zero_state(1)
    assert qf.P0(0).run(ket).norm() == 1.0

    ket = qf.H(0).run(ket)

    measure0 = qf.P0(0).run(ket)
    assert np.isclose(measure0.norm(), 0.5)

    measure1 = qf.P1(0).run(ket)
    assert np.isclose(measure1.norm(), 0.5)


def test_not_unitary() -> None:
    assert not qf.almost_unitary(qf.P0())
    assert not qf.almost_unitary(qf.P1())


# FIXME, add conditional gate
def test_conditional_gate() -> None:
    controlled_gate = qf.conditional_gate(0, qf.X(1), qf.Y(1))

    state = qf.zero_state(2)
    state = controlled_gate.run(state)
    assert state.tensor[0, 1] == 1.0

    state = qf.X(0).run(state)
    state = controlled_gate.run(state)
    assert 1.0j * state.tensor[1, 0] == 1.0


def test_print_gate() -> None:
    stream = io.StringIO()

    qf.print_gate(qf.CNot(0, 1), file=stream)
    s = stream.getvalue()
    ref = (
        "00 -> 00 : (1+0j)\n"
        + "01 -> 01 : (1+0j)\n"
        + "10 -> 11 : (1+0j)\n"
        + "11 -> 10 : (1+0j)\n"
    )

    assert s == ref


def test_inverse_random() -> None:
    K = 4
    for _ in range(REPS):
        gate = qf.RandomGate(range(K))
        inv = gate.H
        gate1 = inv @ gate
        # TODO: almost_identity
        assert qf.gates_close(qf.IdentityGate([0, 1, 2, 3]), gate1)


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import random

import quantumflow as qf

from ..config_test import REPS


def test_XY() -> None:
    gate = qf.XY(-0.1, 0, 1)
    assert qf.gates_close(gate, qf.ISwap(0, 1) ** 0.2)


def test_CV() -> None:
    gate0 = qf.CNot(0, 1) ** 0.5
    gate1 = qf.CV(0, 1)
    assert qf.gates_close(gate0, gate1)

    gate2 = qf.CV_H(0, 1)
    assert qf.gates_close(gate0.H, gate2)

    assert qf.gates_close(gate1.H.H, gate1)


def test_CH() -> None:
    gate1 = qf.CH(0, 1)

    # I picked up this circuit for a CH gate from qiskit
    # qiskit/extensions/standard/ch.py
    # But it clearly far too long. CH is locally equivalent to CNOT,
    # so requires only one CNOT gate.
    circ2 = qf.Circuit(
        [
            qf.H(1),
            qf.S_H(1),
            qf.CNot(0, 1),
            qf.H(1),
            qf.T(1),
            qf.CNot(0, 1),
            qf.T(1),
            qf.H(1),
            qf.S(1),
            qf.X(1),
            qf.S(0),
        ]
    )
    assert qf.gates_close(gate1, circ2.asgate())

    # Here's a better decomposition
    circ1 = qf.Circuit([qf.YPow(+0.25, 1), qf.CNot(0, 1), qf.YPow(-0.25, 1)])
    assert qf.gates_close(gate1, circ1.asgate())
    assert qf.circuits_close(circ1, circ2)


def test_cnot_reverse() -> None:
    # Hadamards reverse control on CNot
    gate0 = qf.H(0) @ qf.IdentityGate([0, 1])
    gate0 = qf.H(1) @ gate0
    gate0 = qf.CNot(1, 0) @ gate0
    gate0 = qf.H(0) @ gate0
    gate0 = qf.H(1) @ gate0

    assert qf.gates_close(qf.CNot(0, 1), gate0)


def test_xy() -> None:
    assert qf.gates_close(qf.XY(0, 0, 1), qf.IdentityGate([0, 1]))
    assert qf.gates_close(qf.XY(-0.5, 0, 1), qf.ISwap(0, 1))
    assert qf.gates_close(qf.XY(0.5, 0, 1), qf.ISwap(0, 1).H)


def test_cv() -> None:
    gate0 = qf.CV(0, 1) ** 2
    assert qf.gates_close(gate0, qf.CNot(0, 1))

    gate1 = qf.CV_H(0, 1) ** 2
    assert qf.gates_close(gate1, qf.CNot(0, 1).H)


def test_CNotPow() -> None:
    q0, q1 = 2, 3
    for _ in range(REPS):
        t = random.uniform(-4, +4)
        gate0 = qf.CNotPow(t, q0, q1)
        gate1 = qf.ControlGate([q0], qf.XPow(t, q1))
        gate2 = qf.CNot(q0, q1) ** t
        gate3 = qf.CNotPow(1, q0, q1) ** t

        assert qf.gates_close(gate0, gate1)
        assert qf.gates_close(gate0, gate2)
        assert qf.gates_close(gate0, gate3)


def test_specialize_2q() -> None:
    q0, q1 = 3, 5
    assert isinstance(qf.Can(0.0, 0.0, 0.0, q0, q1).specialize(), qf.I)
    assert isinstance(qf.Can(0.213, 0.0, 0.0, q0, q1).specialize(), qf.XX)
    assert isinstance(qf.Can(0.0, 0.213, 0.0, q0, q1).specialize(), qf.YY)
    assert isinstance(qf.Can(0.0, 0.0, 0.213, q0, q1).specialize(), qf.ZZ)
    assert isinstance(qf.Can(0.213, 0.213, 0.213, q0, q1).specialize(), qf.Exch)
    assert isinstance(qf.Can(0.5, 0.32, 0.213, q0, q1).specialize(), qf.Can)

    assert isinstance(qf.CNotPow(0.2, q0, q1).specialize(), qf.CNotPow)
    assert isinstance(qf.CNotPow(0.0, q0, q1).specialize(), qf.I)
    assert isinstance(qf.CNotPow(1.0, q0, q1).specialize(), qf.CNot)

    assert isinstance(qf.XX(0.0, q0, q1).specialize(), qf.I)
    assert isinstance(qf.YY(0.0, q0, q1).specialize(), qf.I)
    assert isinstance(qf.ZZ(0.0, q0, q1).specialize(), qf.I)
    assert isinstance(qf.Exch(0.0, q0, q1).specialize(), qf.I)


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import networkx as nx
import numpy as np
import pytest

from quantumflow import utils


def test_cached_property() -> None:
    class Thing:
        def __init__(self, value: int) -> None:
            self.value = value

        @utils.cached_property
        def plus1(self) -> int:
            return self.value + 1

        @utils.cached_property
        def plus2(self) -> int:
            return self.value + 2

    two = Thing(2)
    assert two.plus1 == 2 + 1
    assert two.plus1 == 2 + 1
    assert two.plus2 == 2 + 2
    assert two.plus1 == 2 + 1

    ten = Thing(10)
    assert ten.plus1 == 10 + 1
    assert ten.plus1 == 10 + 1
    assert ten.plus2 == 10 + 2
    assert ten.plus2 == 10 + 2

    assert two.plus1 == 2 + 1
    assert two.plus2 == 2 + 2


def test_deprecated() -> None:
    class Something:
        @utils.deprecated
        def some_thing(self) -> None:
            pass

    obj = Something()

    with pytest.deprecated_call():
        obj.some_thing()


def test_invert_dict() -> None:
    foo = {1: 7, 2: 8, 3: 9}
    bar = utils.invert_map(foo)
    assert bar == {7: 1, 8: 2, 9: 3}

    foo = {1: 7, 2: 8, 3: 7}
    bar = utils.invert_map(foo, one_to_one=False)
    assert bar == {7: set([1, 3]), 8: set([2])}


def test_frozen_dict() -> None:
    f0: utils.FrozenDict[str, int] = utils.FrozenDict({"a": 1, "b": 2})
    assert str(f0) == "FrozenDict({'a': 1, 'b': 2})"

    hash(f0)

    f1 = f0.copy()
    assert f0 == f1

    f2 = f0.update(a=0, c=3)
    assert f2["a"] == 0

    assert "c" in f2
    assert len(f2) == 3

    assert list(f2.keys()) == ["a", "b", "c"]


def test_frozen_dict_generic() -> None:
    f0: utils.FrozenDict[str, int] = utils.FrozenDict({"a": 1, "b": 2})

    k1 = f0["b"]  # noqa

    f1: utils.FrozenDict[str, str] = utils.FrozenDict({"a": 1, "b": 2})  # noqa
    k2: int = f0["b"]  # noqa


def test_bitlist_to_int() -> None:
    assert utils.bitlist_to_int([1, 0, 0]) == 4


def test_int_to_bitlist() -> None:
    assert utils.int_to_bitlist(4, 4) == [0, 1, 0, 0]
    assert utils.int_to_bitlist(4) == [1, 0, 0]


def test_rationalize() -> None:
    flt = 1 / 12
    utils.rationalize(flt)

    with pytest.raises(ValueError):
        utils.rationalize(np.pi)


def test_graph6() -> None:
    graph0 = nx.random_regular_graph(4, 10)
    g0 = utils.to_graph6(graph0)
    graph1 = utils.from_graph6(g0)
    g1 = utils.to_graph6(graph1)

    assert g0 == g1


def test_spanning_tree_count() -> None:
    grp = nx.grid_2d_graph(3, 3)
    count = utils.spanning_tree_count(grp)
    assert count == 192


def test_octagonal_tiling_graph() -> None:
    grp = utils.octagonal_tiling_graph(4, 4)
    assert len(grp) == 128


def test_truncated_grid_2d_graph() -> None:
    G = utils.truncated_grid_2d_graph(12, 11)
    assert len(G) == 72

    G = utils.truncated_grid_2d_graph(8, 8, 2)
    assert len(G) == 8 * 8 - 4 * 3


def test_multi_slice() -> None:
    ms = utils.multi_slice([3, 0], [1, 2])
    assert ms == (2, slice(None, None, None), slice(None, None, None), 1, Ellipsis)


def test_almost_integer() -> None:
    assert utils.almost_integer(1)
    assert utils.almost_integer(1.0)
    assert utils.almost_integer(11239847819349871423)
    assert utils.almost_integer(-4)
    assert utils.almost_integer(-4 + 0j)
    assert utils.almost_integer(1.0000000000000001)
    assert not utils.almost_integer(1.0000001)
    assert not utils.almost_integer(1 + 1j)


# fin



# Copyright 2019-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import random

import numpy as np

import quantumflow as qf

from ..config_test import REPS


def test_CPhase_gates() -> None:
    for _ in range(REPS):
        theta = random.uniform(-4 * np.pi, +4 * np.pi)

        gate11 = qf.ControlGate([0], qf.PhaseShift(theta, 1))
        assert qf.gates_close(gate11, qf.CPhase(theta, 0, 1))

        gate00 = qf.X(0) @ qf.IdentityGate([0, 1])
        gate00 = qf.X(1) @ gate00
        gate00 = gate11 @ gate00
        gate00 = qf.X(0) @ gate00
        gate00 = qf.X(1) @ gate00
        assert qf.gates_close(gate00, qf.CPhase00(theta))

        gate10 = qf.X(0) @ qf.IdentityGate([0, 1])
        gate10 = qf.X(1) @ gate10
        gate10 = qf.CPhase01(theta, 0, 1) @ gate10
        gate10 = qf.X(0) @ gate10
        gate10 = qf.X(1) @ gate10
        assert qf.gates_close(gate10, qf.CPhase10(theta))

        gate0 = qf.CPhase(theta) ** 2
        gate1 = qf.CPhase(theta * 2)
        assert qf.gates_close(gate0, gate1)


def test_CPhase_pow() -> None:
    gate0 = qf.CZ(0, 1) ** 0.4
    gate1 = qf.CPhase(0.4 * np.pi, 0, 1)
    assert qf.gates_close(gate0, gate1)


def test_pswap() -> None:
    assert qf.gates_close(qf.Swap(2, 4), qf.PSwap(0, 2, 4))
    assert qf.gates_close(qf.ISwap(3, 2), qf.PSwap(np.pi / 2, 3, 2))


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import random
from itertools import chain
from typing import Type

import numpy as np
import pytest

import quantumflow as qf
from quantumflow.visualization import kwarg_to_symbol


def _randomize_gate(gatet: Type[qf.StdGate]) -> qf.StdGate:
    """Given a StdGate subclass, return an instance with randomly initialized
    parameters and qubits"""
    args = {arg: random.uniform(-4, 4) for arg in gatet.cv_args}

    qubits = list(range(gatet.cv_qubit_nb))
    random.shuffle(qubits)

    for q in range(gatet.cv_qubit_nb):
        args[f"q{q}"] = qubits[q]

    gate = gatet(**args)  # type: ignore

    return gate


def test_str() -> None:
    g0 = qf.Rx(3.12, 2)
    assert str(g0) == "Rx(3.12) 2"

    g1 = qf.H(0)
    assert str(g1) == "H 0"

    g2 = qf.CNot(0, 1)
    assert str(g2) == "CNot 0 1"


def test_repr() -> None:
    g0 = qf.H(0)
    assert repr(g0) == "H(0)"

    g1 = qf.Rx(3.12, 0)
    assert repr(g1) == "Rx(3.12, 0)"

    g2 = qf.CNot(0, 1)
    assert repr(g2) == "CNot(0, 1)"


@pytest.mark.parametrize("gatet", qf.StdGate.cv_stdgates.values())
def test_stdgates_repr(gatet: Type[qf.StdGate]) -> None:

    gate0 = _randomize_gate(gatet)
    rep = repr(gate0)
    gate1 = eval(rep, qf.StdGate.cv_stdgates)
    assert type(gate0) == type(gate1)
    assert qf.gates_close(gate0, gate1)


@pytest.mark.parametrize("gatet", qf.StdGate.cv_stdgates.values())
def test_stdgates(gatet: Type[qf.StdGate]) -> None:

    # Test creation
    gate = _randomize_gate(gatet)

    # Test correct number of qubits
    assert gate.qubit_nb == gatet.cv_qubit_nb

    # Test hermitian conjugate
    inv_gate = gate.H
    gate.tensor
    inv_gate.tensor

    # Test inverse
    eye = gate @ inv_gate
    assert qf.gates_close(qf.IdentityGate(range(gate.qubit_nb)), eye)
    assert qf.gates_phase_close(qf.IdentityGate(range(gate.qubit_nb)), eye)

    # Test pow
    assert qf.gates_close(gate ** -1, inv_gate)
    assert qf.gates_close((gate ** 0.5) ** 2, gate)
    assert qf.gates_close((gate ** 0.3) @ (gate ** 0.7), gate)

    hgate = qf.Unitary((gate ** 0.5).tensor, gate.qubits)
    assert qf.gates_close(hgate @ hgate, gate)


def _tensor_structure(tensor: qf.QubitTensor) -> "str":
    # "identity", "diagonal", "permutation", "monomial"
    # TODO: Swap

    N = np.ndim(tensor) // 2
    M = np.reshape(tensor, (2 ** N, 2 ** N))

    if np.all(M == np.eye(2 ** N)):
        return "identity"

    if np.all(M == np.diag(np.diagonal(M))):
        return "diagonal"

    if (
        (M.sum(axis=0) == 1).all()
        and (M.sum(axis=0) == 1).all()
        and ((M == 1) | (M == 0)).all()
    ):
        return "permutation"

    A = np.abs(M)
    if (A.sum(axis=0) == A.max(axis=0)).all() and (
        A.sum(axis=1) == A.max(axis=1)
    ).all():
        return "monomial"

    return None


@pytest.mark.parametrize("gatet", qf.StdGate.cv_stdgates.values())
def test_tensor_properties(gatet: Type[qf.StdGate]) -> None:

    gate = _randomize_gate(gatet)

    assert qf.almost_unitary(gate)

    if gatet.cv_hermitian:
        assert gate is gate.H
        assert qf.almost_hermitian(gate)

    structure = gatet.cv_tensor_structure

    assert structure in (None, "identity", "diagonal", "permutation", "monomial")

    assert structure == _tensor_structure(gate.tensor)

    if gate.cv_interchangeable:
        qbs = list(gate.qubits)
        random.shuffle(qbs)
        perm_gate = gate.on(*qbs)
        assert qf.gates_close(gate, perm_gate)


@pytest.mark.parametrize("gatet", qf.StdGate.cv_stdgates.values())
def test_hamiltonians(gatet: Type[qf.StdGate]) -> None:

    gate0 = _randomize_gate(gatet)

    qbs = gate0.qubits
    ham = gate0.hamiltonian

    gate1 = qf.unitary_from_hamiltonian(ham, qbs)

    assert qf.gates_close(gate0, gate1)
    assert qf.gates_phase_close(gate0, gate1)


@pytest.mark.parametrize("gatet", qf.StdGate.cv_stdgates.values())
def test_symbolic(gatet: Type[qf.StdGate]) -> None:
    if len(gatet.cv_args) == 0:
        return
    args = {kwarg_to_symbol[arg]: random.uniform(-4, 4) for arg in gatet.cv_args}
    qbs = range(gatet.cv_qubit_nb)

    # Make gate with symbols
    gate0 = gatet(*chain(args, qbs))
    gate1 = gate0.resolve(subs=args)
    gate2 = gatet(*args.values(), *qbs)  # type: ignore
    assert isinstance(gate1, qf.StdGate)
    assert qf.gates_close(gate1, gate2)

    # Arguments with symbolic constants should be converted to float before
    # creating tensor
    gate0 = gatet(*chain(([qf.PI] * len(args)), qbs))
    gate0.tensor


@pytest.mark.parametrize("gatet", qf.StdGate.cv_stdgates.values())
def test_decompose(gatet: Type[qf.StdGate]) -> None:
    gate0 = _randomize_gate(gatet)
    circ = qf.Circuit(gate0.decompose())
    gate1 = circ.asgate()
    assert qf.gates_close(gate0, gate1)


# fin
